import numpy as np
import pytest

pytest.importorskip("gymnasium")

from citylearn.base import EpisodeTracker
from citylearn.data import ChargerSimulation
from citylearn.energy_model import Battery
from citylearn.electric_vehicle import ElectricVehicle
from citylearn.electric_vehicle_charger import Charger


def _make_tracker(length: int = 3) -> EpisodeTracker:
    tracker = EpisodeTracker(0, length - 1)
    tracker.next_episode(length, False, False, 0)
    return tracker


def _make_simulation(length: int, ev_id: str) -> ChargerSimulation:
    return ChargerSimulation(
        electric_vehicle_charger_state=[1] * length,
        electric_vehicle_id=[ev_id] * length,
        electric_vehicle_departure_time=[-1] * length,
        electric_vehicle_required_soc_departure=[-0.1] * length,
        electric_vehicle_estimated_arrival_time=[-1] * length,
        electric_vehicle_estimated_soc_arrival=[-0.1] * length,
    )


def _make_ev(tracker: EpisodeTracker, initial_soc: float = 0.5) -> ElectricVehicle:
    battery = Battery(
        capacity=100.0,
        nominal_power=50.0,
        initial_soc=initial_soc,
        efficiency=1.0,
        capacity_loss_coefficient=0.0,
        power_efficiency_curve=[[0.0, 1.0], [1.0, 1.0]],
        capacity_power_curve=[[0.0, 1.0], [1.0, 1.0]],
        loss_coefficient=0.0,
        seconds_per_time_step=3600,
        episode_tracker=tracker,
    )
    ev = ElectricVehicle(episode_tracker=tracker, battery=battery, seconds_per_time_step=3600)
    ev.reset()
    return ev


def _make_charger(tracker: EpisodeTracker, sim: ChargerSimulation, ev: ElectricVehicle, **kwargs) -> Charger:
    charger = Charger(
        episode_tracker=tracker,
        charger_simulation=sim,
        charger_id="test",
        max_charging_power=kwargs.get("max_charging_power", 10.0),
        max_discharging_power=kwargs.get("max_discharging_power", 10.0),
        efficiency=kwargs.get("efficiency", 1.0),
        connected_electric_vehicle=ev,
        seconds_per_time_step=3600,
    )
    charger.reset()
    charger.connected_electric_vehicle = ev
    return charger


def test_charger_charging_updates_ev_soc_and_accounting():
    tracker = _make_tracker()
    ev = _make_ev(tracker)
    sim = _make_simulation(3, ev.name)
    charger = _make_charger(tracker, sim, ev, efficiency=0.9)

    initial_soc = ev.battery.soc[ev.time_step]
    charger.update_connected_electric_vehicle_soc(0.5)

    energy_balance = ev.battery.energy_balance[ev.time_step]
    electricity_consumption = charger.electricity_consumption[charger.time_step]

    expected_soc = (initial_soc * ev.battery.capacity + energy_balance) / ev.battery.capacity

    assert np.isclose(ev.battery.soc[ev.time_step], expected_soc)
    assert np.isclose(energy_balance, electricity_consumption * charger.efficiency)


def test_charger_discharging_updates_ev_soc_and_accounting():
    tracker = _make_tracker()
    ev = _make_ev(tracker)
    sim = _make_simulation(3, ev.name)
    charger = _make_charger(tracker, sim, ev, efficiency=0.85)

    initial_soc = ev.battery.soc[ev.time_step]
    charger.update_connected_electric_vehicle_soc(-0.4)

    energy_balance = ev.battery.energy_balance[ev.time_step]
    electricity_consumption = charger.electricity_consumption[charger.time_step]

    expected_soc = (initial_soc * ev.battery.capacity + energy_balance) / ev.battery.capacity

    assert np.isclose(ev.battery.soc[ev.time_step], expected_soc)
    assert np.isclose(energy_balance, electricity_consumption / charger.efficiency)


def test_charger_action_without_connected_vehicle_has_no_effect():
    tracker = _make_tracker()
    ev = _make_ev(tracker)
    sim = _make_simulation(3, ev.name)
    charger = _make_charger(tracker, sim, ev)

    charger.connected_electric_vehicle = None
    charger.update_connected_electric_vehicle_soc(0.8)

    assert np.allclose(charger.electricity_consumption, 0.0)
    expected_commanded_energy = 0.8 * charger.max_charging_power * charger.algorithm_action_based_time_step_hours_ratio
    assert np.isclose(
        charger.past_charging_action_values_kwh[charger.time_step],
        expected_commanded_energy,
    )



import concurrent.futures
from copy import deepcopy
import os
from enum import Enum, unique
from pathlib import Path
import random
import shutil
from typing import Any, Callable, List, Mapping, Tuple, Union
from doe_xstock.data import VersionDatasetType
from doe_xstock.end_use_load_profiles import EndUseLoadProfiles
from doe_xstock.simulate import EndUseLoadProfilesEnergyPlusSimulator, OpenStudioModelEditor
import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_percentage_error
import torch
from citylearn.agents.base import Agent, BaselineAgent
from citylearn.base import Environment
from citylearn.building import Building
from citylearn.citylearn import CityLearnEnv
from citylearn.data import get_settings
from citylearn.dynamics import LSTMDynamics
from citylearn.end_use_load_profiles.clustering import MetadataClustering
from citylearn.end_use_load_profiles.lstm_model.model_generation_wrapper import run_one_model
from citylearn.end_use_load_profiles.simulate import EndUseLoadProfilesEnergyPlusPartialLoadSimulator
from citylearn.preprocessing import PeriodicNormalization, Normalize
from citylearn.utilities import FileHandler

BuildingSimulators = Mapping[
    str, 
    Union[EndUseLoadProfilesEnergyPlusSimulator, EndUseLoadProfilesEnergyPlusSimulator, List[EndUseLoadProfilesEnergyPlusPartialLoadSimulator]]
]
BuildingsSimulators = List[BuildingSimulators]

@unique
class SampleMethod(Enum):
    RANDOM = 0
    METADATA_CLUSTER_FREQUENCY = 1

class NeighborhoodBuild:
    def __init__(
        self, schema_filepath: Path, citylearn_simulation_test_evaluation: pd.DataFrame, citylearn_simulation_lstm_prediction_data: Mapping[int, pd.DataFrame], 
        citylearn_simulation_lstm_error_data: Mapping[int, pd.DataFrame], lstm_test_data: Mapping[int, pd.DataFrame], bldg_ids: List[int], sample_cluster_labels: List[int], 
        sample_metadata: Mapping[str, Any], simulators: BuildingsSimulators
    ):
        self.__schema_filepath = schema_filepath
        self.__citylearn_simulation_test_evaluation = citylearn_simulation_test_evaluation
        self.__citylearn_simulation_lstm_prediction_data = citylearn_simulation_lstm_prediction_data
        self.__citylearn_simulation_lstm_error_data = citylearn_simulation_lstm_error_data
        self.__lstm_test_data = lstm_test_data
        self.__bldg_ids = bldg_ids
        self.__sample_cluster_labels = sample_cluster_labels
        self.__sample_metadata = sample_metadata
        self.__simulators = simulators

    @property
    def schema_filepath(self) -> Path:
        return self.__schema_filepath

    @property
    def citylearn_simulation_test_evaluation(self) -> pd.DataFrame:
        return self.__citylearn_simulation_test_evaluation
    
    @property
    def citylearn_simulation_lstm_prediction_data(self) -> List[pd.DataFrame]:
        return self.__citylearn_simulation_lstm_prediction_data
    
    @property
    def citylearn_simulation_lstm_error_data(self) -> List[pd.DataFrame]:
        return self.__citylearn_simulation_lstm_error_data
    
    @property
    def lstm_test_data(self) -> List[pd.DataFrame]:
        return self.__lstm_test_data
    
    @property
    def bldg_ids(self) -> List[int]:
        return self.__bldg_ids
    
    @property
    def sample_cluster_labels(self) -> List[int]:
        return self.__sample_cluster_labels
    
    @property
    def sample_metadata(self) -> Mapping[str, Any]:
        return self.__sample_metadata
    
    @property
    def simulators(self) -> BuildingsSimulators:
        return self.__simulators

class Neighborhood:
    __BUILDING_TYPE_COLUMN = 'in.geometry_building_type_recs'
    __SINGLE_FAMILY_BUILDING_TYPE_NAME = 'Single-Family Detached'
    __COUNTY_COLUMN = 'in.resstock_county_id'

    def __init__(
        self, weather_data: str = None, year_of_publication: int = None, release: int = None, cache: bool = None,
        energyplus_output_directory: Union[Path, str] = None, dataset_directory: Union[Path, str] = None,
        max_workers: int = None, random_seed: int = None
    ) -> None:
        self.__end_use_load_profiles = EndUseLoadProfiles(
            dataset_type=VersionDatasetType.RESSTOCK,
            weather_data=weather_data,
            year_of_publication=year_of_publication,
            release=release,
            cache=cache,
        )
        self.energyplus_output_directory = energyplus_output_directory
        self.dataset_directory = dataset_directory
        self.max_workers = max_workers
        self.random_seed = random_seed

    @property
    def end_use_load_profiles(self) -> EndUseLoadProfiles:
        return self.__end_use_load_profiles
    
    @property
    def energyplus_output_directory(self) -> Union[Path, str]:
        return self.__energyplus_output_directory
    
    @property
    def dataset_directory(self) -> Union[Path, str]:
        return self.__dataset_directory
    
    @property
    def max_workers(self) -> int:
        return self.__max_workers
    
    @property
    def random_seed(self) -> int:
        return self.__random_seed

    @energyplus_output_directory.setter
    def energyplus_output_directory(self, value: Union[Path, str]):
        self.__energyplus_output_directory = 'neighborhood_energyplus_output' if value is None else value

    @dataset_directory.setter
    def dataset_directory(self, value: Union[Path, str]):
        self.__dataset_directory = 'neighborhood_datasets' if value is None else value

    @max_workers.setter
    def max_workers(self, value: int):
        self.__max_workers = value

    @random_seed.setter
    def random_seed(self, value: int):
        self.__random_seed = random.randint(*Environment.DEFAULT_RANDOM_SEED_RANGE) if value is None else value

    def build(
        self, idd_filepath: Union[Path, str], bldg_ids: List[int] = None, include_lstm_models: bool = None, test_lstm_models: bool = None, 
        test_citylearn_simulation: bool = None, delete_energyplus_simulation_output: bool = None, sample_buildings_kwargs: Mapping[str, Any] = None, 
        energyplus_simulation_kwargs: Mapping[str, Any] = None, train_lstm_kwargs: Mapping[str, Any] = None, schema_kwargs: Mapping[str, Any] = None, 
        test_citylearn_simulation_kwargs: Mapping[str, Any] = None
    ) -> NeighborhoodBuild:
        sample_buildings_kwargs = {} if sample_buildings_kwargs is None else sample_buildings_kwargs
        energyplus_simulation_kwargs = {} if energyplus_simulation_kwargs is None else energyplus_simulation_kwargs
        train_lstm_kwargs = {} if train_lstm_kwargs is None else train_lstm_kwargs
        schema_kwargs = {} if schema_kwargs is None else schema_kwargs
        test_citylearn_simulation_kwargs = {} if test_citylearn_simulation_kwargs is None else test_citylearn_simulation_kwargs
        include_lstm_models = True if include_lstm_models is None else include_lstm_models
        test_citylearn_simulation = True if test_citylearn_simulation is None else test_citylearn_simulation
        test_lstm_models = True if test_lstm_models is None else test_lstm_models
        delete_energyplus_simulation_output = False if delete_energyplus_simulation_output is None else delete_energyplus_simulation_output

        labels = None
        sample_metadata = None
        simulators = None
        lstm_training_data = None
        lstm_models = None
        schema_filepath = None
        citylearn_simulation_evaluation = None 
        citylearn_simulation_lstm_prediction_data = None
        citylearn_simulation_lstm_error_data = None
        lstm_test_data = None

        # sample buildings
        if bldg_ids is None:
            bldg_ids, labels, sample_metadata = self.sample_buildings(**sample_buildings_kwargs)
        
        else:
            pass
        
        # run energyplus simulations
        simulators = self.simulate_energy_plus(bldg_ids, idd_filepath, **energyplus_simulation_kwargs)

        # train LSTM models
        if include_lstm_models:
            lstm_training_data = self.get_lstm_training_data(simulators)
            lstm_models = self.train_lstm(lstm_training_data, **train_lstm_kwargs)
        
        else:
            pass
        
        # generate schema
        schema_filepath = self.set_schema(simulators, bldg_ids, lstm_models=lstm_models, **schema_kwargs)

        # confirm schema initializes correctly in CityLearn and simulation completes without errors
        if test_citylearn_simulation:
            citylearn_simulation_evaluation, citylearn_simulation_lstm_prediction_data, citylearn_simulation_lstm_error_data \
            = self.test_citylearn_simulation(schema_filepath, **test_citylearn_simulation_kwargs)
        
        else:
            pass
        
        # run test on generated LSTM training data for all partial load simulation references
        if include_lstm_models and test_lstm_models:
            lstm_test_data = self.multiprocess_test_lstm(lstm_training_data, schema_filepath)
        
        else:
            pass
        
        # delete energyplus output files to save space
        if delete_energyplus_simulation_output:
            self.delete_energyplus_simulation_output(simulators)
        
        else:
            pass

        return NeighborhoodBuild(
            schema_filepath=schema_filepath,
            citylearn_simulation_test_evaluation=citylearn_simulation_evaluation,
            citylearn_simulation_lstm_prediction_data=citylearn_simulation_lstm_prediction_data,
            citylearn_simulation_lstm_error_data=citylearn_simulation_lstm_error_data,
            lstm_test_data=lstm_test_data,
            bldg_ids=bldg_ids,
            sample_cluster_labels=labels,
            sample_metadata=sample_metadata,
            simulators=simulators
        )
    
    def delete_energyplus_simulation_output(self, simulators: BuildingsSimulators):
        directories = [Path(s['ideal'].output_directory).parents[0] for s in simulators]

        for d in directories:
            if os.path.isdir(d):
                shutil.rmtree(d)
            
            else:
                pass

    def multiprocess_test_lstm(self, training_data: List[pd.DataFrame], schema_filepath: Path) -> List[pd.DataFrame]:
        test_data = [None]*len(training_data)

        with concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers) as executor:
            results = [executor.submit(
                self.test_lstm, 
                *(i, td.copy(), schema_filepath)
            ) for i, td in enumerate(training_data)]
            
            for future in concurrent.futures.as_completed(results):
                result = future.result()
                test_data[result[0]] = result[1]

        return test_data
    
    def test_lstm(self, bldg_ix: int, training_data: pd.DataFrame, schema_filepath: Path) -> Tuple[int, pd.DataFrame]:
        # set data
        schema = FileHandler.read_json(schema_filepath)
        schema_directory = Path(schema_filepath).parents[0]
        bldg_key = list(schema['buildings'].keys())[bldg_ix]
        dynamics_model = schema['buildings'][bldg_key]['dynamics']
        normalization_minimum = dynamics_model['attributes']['input_normalization_minimum']
        normalization_maximum = dynamics_model['attributes']['input_normalization_maximum']
        observation_names = dynamics_model['attributes']['input_observation_names']
        periodic_observations = Building.get_periodic_observation_metadata()
        training_data = training_data.sort_values(['reference', 'timestep'])

        # type cast
        columns = [c for c in observation_names + list(periodic_observations.keys()) if not c.endswith('_sin') and not c.endswith('_cos')]
        training_data[columns] = training_data[columns].astype('float32')

        ## cyclic normalization
        for c, m in periodic_observations.items():
            result = training_data[c]*PeriodicNormalization(x_max=m[-1])
            result = pd.DataFrame(result.tolist(), index=result.index)
            training_data[f'{c}_sin'] = result[0].tolist()
            training_data[f'{c}_cos'] = result[1].tolist()

        input_columns = []
        
        # then normalize all columns between 0 and 1 including the cyclic columns
        for i, c in enumerate(observation_names):
            input_columns.append(f'{c}_norm')
            training_data[input_columns[-1]] = training_data[c]*Normalize(normalization_minimum[i], normalization_maximum[i])

        training_data[input_columns] = training_data[input_columns].astype('float32')
        independent_columns = input_columns[:-1]
        dependent_column = input_columns[-1]

        # initialize trained model
        model = LSTMDynamics(
            os.path.join(schema_directory, dynamics_model['attributes']['filename']),
            observation_names,
            normalization_minimum,
            normalization_maximum,
            hidden_size=dynamics_model['attributes']['hidden_size'],
            num_layers=dynamics_model['attributes']['num_layers'],
            lookback=dynamics_model['attributes']['lookback']
        )

        # make predictions
        data_list = []

        for _, reference_data in training_data.groupby('reference'):
            reference_data = reference_data.reset_index(drop=True)
            model.reset()

            for i in range(model.lookback, reference_data.shape[0]):
                x_independent = reference_data[independent_columns].iloc[i - model.lookback + 1:i + 1].values
                x_dependent = reference_data[[dependent_column]].iloc[i - model.lookback:i].values
                x = np.append(x_independent, x_dependent, axis=1)
                x = torch.tensor(x.copy())
                x = x[np.newaxis, :, :]
                hidden_state = tuple([h.data for h in model._hidden_state])
                y, model._hidden_state = model(x.float(), hidden_state)
                y = y.item()
                reference_data.loc[i, dependent_column] = y

            predicted_column = dependent_column.replace('norm', 'predicted')
            actual_column = dependent_column.replace('_norm', '')
            norm_min = normalization_minimum[-1]
            norm_max = normalization_maximum[-1]
            reference_data[predicted_column] = reference_data[dependent_column]*(norm_max - norm_min) + norm_min
            reference_data = reference_data[['timestep', 'reference', 'reference_name', actual_column, predicted_column]].copy()
            data_list.append(reference_data)

        return bldg_ix, pd.concat(data_list, ignore_index=True)

    def test_citylearn_simulation(
        self, schema: Path, model: Agent = None, env_kwargs: Mapping[str, Any] = None, model_kwargs: Mapping[str, Any] = None,
          report_lstm_performance: bool = None
    ) -> Tuple[pd.DataFrame, List[pd.DataFrame], List[pd.DataFrame]]:
        report_lstm_performance = True if report_lstm_performance is None else report_lstm_performance
        env_kwargs = {} if env_kwargs is None else env_kwargs
        model_kwargs = {} if model_kwargs is None else model_kwargs
        
        # test initialization
        env = CityLearnEnv(schema, **env_kwargs)
        model = BaselineAgent if model is None else model
        model = model(env=env, **model_kwargs)
        observations, _ = env.reset()
        
        # test stepping
        while not env.terminated:
            actions = model.predict(observations)
            observations, _, _, _, _ = env.step(actions)

        evaluation = env.evaluate()

        if report_lstm_performance:
            lstm_prediction_data: List[pd.DataFrame] = []
            lstm_error_data: List[pd.DataFrame] = []

            for b in env.buildings:
                lstm_prediction_data.append(pd.DataFrame({
                    'hour': b.energy_simulation.hour,
                    'month': b.energy_simulation.month,
                    'indoor_dry_bulb_temperature': b.energy_simulation.indoor_dry_bulb_temperature_without_control,
                    'indoor_dry_bulb_temperature_predicted': b.indoor_dry_bulb_temperature,
                }))
                
                rmse_data = lstm_prediction_data[-1].groupby(['month'])[['indoor_dry_bulb_temperature','indoor_dry_bulb_temperature_predicted']].apply(
                    lambda x: mean_squared_error(x['indoor_dry_bulb_temperature'], x['indoor_dry_bulb_temperature_predicted']
                )).reset_index(name='rmse')
                rmse_data['rmse'] = rmse_data['rmse']**0.5
                mape_data = lstm_prediction_data[-1].groupby(['month'])[['indoor_dry_bulb_temperature','indoor_dry_bulb_temperature_predicted']].apply(
                    lambda x: mean_absolute_percentage_error(x['indoor_dry_bulb_temperature'], x['indoor_dry_bulb_temperature_predicted']
                )).reset_index(name='mape')
                mape_data['mape'] = mape_data['mape']*100.0
                error_data = rmse_data.merge(mape_data, on=['month'])
                error_data.insert(0, 'bldg_name', b.name)
                lstm_error_data.append(error_data)
        
        else:
            lstm_prediction_data = None
            lstm_error_data = None


        return evaluation, lstm_prediction_data, lstm_error_data

    def set_schema(
        self, simulators: BuildingsSimulators, bldg_ids: List[int], lstm_models: List[Mapping[str, Any]] = None, 
        pricing: Union[List[Union[Path, pd.DataFrame, str]], Path, pd.DataFrame, str] = None, carbon_intensity: List[Union[Path, pd.DataFrame, str]] = None,
        template: Mapping[str, Union[dict, float, int, str]] = None, 
        metadata: pd.DataFrame = None, dataset_name: str = None, schema_directory: Union[Path, str] = None, weather_kwargs: dict = None,
        user_customization_function: Callable[[dict], dict] = None
    ) -> Path:
        assert pricing is None or not isinstance(pricing, list) or len(pricing) == len(bldg_ids), \
            'There must be as many pricing as bldg_ids.'
        assert carbon_intensity is None or not isinstance(carbon_intensity, list) or len(carbon_intensity) == len(bldg_ids), \
            'There must be as many carbon_intensity as bldg_ids.'
        template = get_settings()['schema']['template'] if template is None else template
        lstm_models = [None]*len(simulators) if lstm_models is None else lstm_models
        metadata = self.end_use_load_profiles.metadata.metadata.get().to_dict('index') if metadata is None else metadata
        schema_directory = self.dataset_directory if schema_directory is None else schema_directory
        county = metadata[list(metadata.keys())[0]][self.__COUNTY_COLUMN]
        county = county.lower().replace(',', '').replace(' ', '_')
        dataset_name = f'{self.end_use_load_profiles.version}-{county}' if dataset_name is None else dataset_name
        weather_kwargs = {} if weather_kwargs is None else weather_kwargs

        building_template = template.pop('buildings')['Building_1']
        template['buildings'] = {}
        schema_directory = os.path.join(schema_directory, dataset_name)
        os.makedirs(schema_directory, exist_ok=True)
        
        # write weather (csv and epw)
        reference_simulator = simulators[0]['partial'][0]
        weather_data = self.get_weather_data(reference_simulator, **weather_kwargs)
        weather_filename = building_template['weather'] if building_template['weather'] is not None else 'weather.csv'
        weather_data.to_csv(os.path.join(schema_directory, weather_filename), index=False)
        _ = shutil.copy(reference_simulator.epw_filepath, os.path.join(schema_directory, weather_filename.replace('.csv', '.epw')))

        for i, (bldg_id, building_simulators, lstm_model) in enumerate(zip(bldg_ids, simulators, lstm_models)):
            bldg_key = '-'.join(building_simulators['ideal'].simulation_id.split('-')[:-1])
            simulator: EndUseLoadProfilesEnergyPlusPartialLoadSimulator = building_simulators['partial'][0]
            ideal_simulator: EndUseLoadProfilesEnergyPlusSimulator = building_simulators['ideal']
            building = deepcopy(building_template)
            no_lstm_model = True if lstm_model is None else False

            # set building data
            building_data_query_filepath = os.path.join(simulator.QUERIES_DIRECTORY, 'select_citylearn_energy_simulation.sql')
            building_data = simulator.get_output_database().query_table_from_file(building_data_query_filepath)
            building_data_ideal = ideal_simulator.get_output_database().query_table_from_file(building_data_query_filepath)
            building_data['indoor_dry_bulb_temperature_cooling_set_point'] = building_data_ideal['indoor_dry_bulb_temperature_cooling_set_point'].tolist()
            building_data['indoor_dry_bulb_temperature_heating_set_point'] = building_data_ideal['indoor_dry_bulb_temperature_heating_set_point'].tolist()

            # set building-specific filepaths
            building['energy_simulation'] = f'{bldg_key}.csv'
            building['pricing'], building['carbon_intensity'] = self.__get_building_pricing_and_carbon_intensity(
                bldg_key,
                i,
                building_template,
                schema_directory,
                pricing,
                carbon_intensity)
            
            # as-modeled DER survey
            no_space_cooling = building_data['cooling_demand'].sum() == 0
            no_space_heating = building_data['heating_demand'].sum() == 0
            no_dhw_heating = building_data['dhw_demand'].sum() == 0
            no_electric_dhw_heating = metadata[bldg_id]['in.water_heater_fuel'] != 'Electricity'
            no_dhw_heating_storage = True if no_electric_dhw_heating else False

            if no_space_cooling:
                building['cooling_device'] = None
                building['inactive_observations'] += [o for o in template['observations'] if 'cooling' in o]

            else:
                pass

            if no_space_heating:
                building['heating_device'] = None
                building['inactive_observations'] += [o for o in template['observations'] if 'heating' in o]

            else:
                pass

            if no_dhw_heating:
                building['dhw_device'] = None
                building['dhw_storage'] = None
                building['inactive_observations'] += [o for o in template['observations'] if 'dhw' in o]
                building['inactive_actions'] += ['dhw_storage']

            elif no_dhw_heating_storage:
                building['dhw_storage'] = None
                building['inactive_observations'] += [o for o in template['observations'] if 'dhw_storage' in o]
                building['inactive_actions'] += ['dhw_storage']

            else:
                pass

            # PV autosize attributes
            osm = OpenStudioModelEditor(self.end_use_load_profiles.get_building(bldg_id).open_studio_model.get())
            building['pv']['autosize_attributes'] = {
                **building['pv']['autosize_attributes'],
                'roof_area': round(osm.get_roof_area(), 2),
            }

            # dynamics model
            if no_lstm_model:
                building['dynamics'] = None
                building['type'] = 'citylearn.building.Building'
            
            else:
                building['dynamics']['attributes']['filename'] = f'{bldg_key}.pth'
                building['dynamics']['attributes']['hidden_size'] = lstm_model['attributes']['hidden_size']
                building['dynamics']['attributes']['num_layers'] = lstm_model['attributes']['num_layers']
                building['dynamics']['attributes']['lookback'] = lstm_model['attributes']['lookback']
                building['dynamics']['attributes']['input_observation_names'] = lstm_model['attributes']['input_observation_names']
                building['dynamics']['attributes']['input_normalization_minimum'] = lstm_model['attributes']['input_normalization_minimum']
                building['dynamics']['attributes']['input_normalization_maximum'] = lstm_model['attributes']['input_normalization_maximum']
                torch.save(
                    lstm_model['model'].state_dict(), 
                    os.path.join(schema_directory, building['dynamics']['attributes']['filename'])
                )

            # set building schema
            template['buildings'][bldg_key] = building

            # write back data file
            building_data = building_data.astype('float32')

            for c in ['hour', 'month', 'daylight_savings_status', 'hvac_mode', 'day_type']:
                if c in building_data.columns:
                    building_data[c] = building_data[c].astype('int32')
                
                else:
                    pass

            building_data.to_csv(os.path.join(schema_directory, building['energy_simulation']), index=False)

        schema_filepath = os.path.join(schema_directory, f'schema.json')
        template = user_customization_function(template) if user_customization_function is not None else template
        FileHandler.write_json(schema_filepath, template)
        
        return schema_filepath
    
    def __get_building_pricing_and_carbon_intensity(
        self, bldg_key: str, bldg_ix: int, 
        building_template: Mapping[str, Union[dict, float, int, str]], schema_directory: Union[str, Path],
        pricing: Union[List[Union[Path, pd.DataFrame, str]], Path, pd.DataFrame, str] = None,
        carbon_intensity: Union[List[Union[Path, pd.DataFrame, str]], Path, pd.DataFrame, str] = None
    ) -> Tuple[str, str]:
        data = dict(
            pricing=dict(
                default_filename=building_template['pricing'] if building_template['pricing'] is not None else 'pricing.csv',
                data=pricing,
                schema_filename=None
            ),
            carbon_intensity=dict(
                default_filename=building_template['carbon_intensity'] if building_template['carbon_intensity'] is not None else 'carbon_intensity.csv',
                data=carbon_intensity,
                schema_filename=None
            )
        )

        for k, v in data.items():
            if v['data'] is None:
                if os.path.isfile(os.path.join(schema_directory, v['default_filename'])):
                    data[k]['schema_filename'] = v['default_filename']

                else:
                    pass

            elif isinstance(v['data'], (Path, str)):
                destination_filepath = os.path.join(schema_directory, v['default_filename'])
                shutil.copy(v['data'], destination_filepath)
                data[k]['schema_filename'] = v['default_filename']

            elif isinstance(v['data'], pd.DataFrame):
                destination_filepath = os.path.join(schema_directory, v['default_filename'])
                v['data'].to_csv(destination_filepath, index=False)
                data[k]['schema_filename'] = v['default_filename']

            elif isinstance(v['data'], list):
                destination_filename = f'{bldg_key}-pricing.csv'
                data[k]['schema_filename'] = destination_filename
                destination_filepath = os.path.join(schema_directory, destination_filename)
                
                if isinstance(v['data'][bldg_ix], (Path, str)):
                    shutil.copy(v['data'][bldg_ix], destination_filepath)

                elif isinstance(v['data'][bldg_ix], pd.DataFrame):
                    v['data'][bldg_ix].to_csv(destination_filepath, index=False)

            else:
                raise ValueError(f'Invalid {k} value type: {type(v["data"])}')

        return data['pricing']['schema_filename'], data['carbon_intensity']['schema_filename']

    def get_weather_data(self, simulator: EndUseLoadProfilesEnergyPlusPartialLoadSimulator, shifts: Tuple[int, int, int] = None, accuracy: Mapping[str, Tuple[float, float, float]] = None) -> pd.DataFrame:
        database = simulator.get_output_database()
        query_filepath = os.path.join(simulator.QUERIES_DIRECTORY, 'select_citylearn_weather.sql')
        data = database.query_table_from_file(query_filepath)
        columns = data.columns
        shifts = (6, 12, 24) if shifts is None else shifts
        accuracy = {c: (0.3, 0.65, 1.35) if c == 'outdoor_dry_bulb_temperature' else (0.025, 0.05, 0.1) for c in columns} \
            if accuracy is None else accuracy

        for c in columns:
            for i, (s, a) in enumerate(zip(shifts, accuracy[c])):
                arr = np.roll(data[c], shift=-s)
                nprs = np.random.RandomState(self.random_seed)
                shift_column = f'{c}_predicted_{int(i + 1)}'

                if c in ['outdoor_dry_bulb_temperature']:
                    data[shift_column] = arr + nprs.uniform(-a, a, len(arr))

                elif c in ['outdoor_relative_humidity', 'diffuse_solar_irradiance', 'direct_solar_irradiance']:
                    data[shift_column] = arr + arr*nprs.uniform(-a, a, len(arr))

                else:
                    raise Exception(f'Unknown field: {c}')
                
                if c != 'outdoor_dry_bulb_temperature':
                    data[shift_column] = data[shift_column].clip(lower=0.0)

                    if c == 'outdoor_relative_humidity':
                        data[shift_column] = data[shift_column].clip(upper=100.0)
                    
                    else:
                        pass

                else:
                    pass

        data = data.astype('float32')

        return data
    
    def train_lstm(self, data: List[pd.DataFrame], config: Mapping[str, Any] = None, seed: int = None) -> List[Mapping[str, Any]]:
        seed = self.random_seed if seed is None else seed
        config = get_settings()['lstm']['train']['config'] if config is None else config
        data = [run_one_model(config, d.copy(), seed) for d in data]

        return data
    
    def get_lstm_training_data(self, simulators: BuildingsSimulators) -> List[pd.DataFrame]:
        data = []

        for building_simulators in simulators:
            data_list = []

            for partial_simulator in building_simulators['partial']:
                partial_simulator: EndUseLoadProfilesEnergyPlusPartialLoadSimulator
                query_filepath = os.path.join(EndUseLoadProfilesEnergyPlusPartialLoadSimulator.QUERIES_DIRECTORY, 'select_lstm_training_data.sql')
                pdata = partial_simulator.get_output_database().query_table_from_file(query_filepath)
                pdata.insert(0, 'reference_name', partial_simulator.simulation_id.split('-')[-2])
                pdata.insert(0, 'reference', int(partial_simulator.simulation_id.split('-')[-1]))
                data_list.append(pdata)
            
            data.append(pd.concat(data_list, ignore_index=True))

        return data
    
    def simulate_energy_plus(
        self, bldg_ids: List[int], idd_filepath: Union[Path, str], simulation_ids: List[str] = None, models: List[Union[Path, str]] = None, 
        schedules: List[Union[Path, pd.DataFrame]] = None, thermostat_setpoints: List[Union[Path, pd.DataFrame]] = None, 
        fixed_ideal_loads_idf: List[Union[Path, str]] = None, osm: bool = None, partial_loads_simulations: int = None, simulate_mechanical: bool = None, 
        delete_existing_simulation: bool = None, partial_loads_kwargs: Mapping[str, Any] = None, **kwargs
    ) -> BuildingsSimulators:
        assert models is None or len(models) == len(bldg_ids), 'There must be as many models as bldg_ids.'
        assert schedules is None or len(schedules) == len(bldg_ids), 'There must be as many schedules as bldg_ids.'
        assert thermostat_setpoints is None or len(thermostat_setpoints) == len(bldg_ids), 'There must be as many thermostat_setpoints as bldg_ids.'
        assert fixed_ideal_loads_idf is None or len(fixed_ideal_loads_idf) == len(bldg_ids), 'There must be as many fixed_ideal_loads_idf as bldg_ids.'
        assert simulation_ids is None or len(simulation_ids) == len(bldg_ids), 'There must be as many simulation_ids as bldg_ids.'
        os.makedirs(self.energyplus_output_directory, exist_ok=True)
        partial_loads_simulations = 4 if partial_loads_simulations is None else partial_loads_simulations
        simulate_mechanical = True if simulate_mechanical is None else simulate_mechanical
        delete_existing_simulation = True if delete_existing_simulation is None else delete_existing_simulation
        partial_loads_kwargs = {} if partial_loads_kwargs is None else partial_loads_kwargs
        simulators = []

        kwargs = dict(
            idd_filepath=idd_filepath,
            osm=osm,
            **kwargs,
            number_of_time_steps_per_hour=1 if kwargs.get('number_of_time_steps_per_hour', None) is None else kwargs['number_of_time_steps_per_hour']
        )

        if (
            kwargs.get('default_output_variables', None) is None or not kwargs['default_output_variables']
        ) and kwargs.get('output_variables', None) is None:
            kwargs['default_output_variables'] = True

        else:
            pass

        if (
            kwargs.get('default_output_meters', None) is None or not kwargs['default_output_meters']
        ) and kwargs.get('output_meters', None) is None:
            kwargs['default_output_meters'] = True

        else:
            pass

        for i, bldg_id in enumerate(bldg_ids):
            bldg_id_ix = bldg_ids[:i].count(bldg_id)
            simulation_id = f'{self.end_use_load_profiles.version}-{bldg_id}-{bldg_id_ix}' if simulation_ids is None else simulation_ids[i]
            output_directory = os.path.join(self.energyplus_output_directory, simulation_id)

            if os.path.isdir(output_directory) and delete_existing_simulation:
                shutil.rmtree(output_directory)
            
            else:
                pass
            
            os.makedirs(output_directory)

            _kwargs = dict(
                **kwargs,
                bldg_id=bldg_id,
                model=models[i] if models is not None else models,
                schedules=schedules[i] if schedules is not None else schedules,
                thermostat_setpoint=thermostat_setpoints[i] if thermostat_setpoints is not None else thermostat_setpoints,
            )

            # mechanical loads
            if simulate_mechanical:
                _simulation_id = f'{simulation_id}-mechanical'
                _output_directory = os.path.join(output_directory, _simulation_id)
                mechanical_simulator = self.end_use_load_profiles.simulate_building(
                    ideal_loads=False,
                    simulation_id=_simulation_id,
                    output_directory=_output_directory,
                    **_kwargs
                ).simulator
            
            else:
                mechanical_simulator = None

            # ideal loads
            _ideal_kwargs = deepcopy(_kwargs)

            if fixed_ideal_loads_idf is not None:
                _ideal_kwargs['model'] = fixed_ideal_loads_idf[i]
                _ideal_kwargs['osm'] = False
            
            else:
                pass
            
            _simulation_id = f'{simulation_id}-ideal'
            _output_directory = os.path.join(output_directory, _simulation_id)
            ideal_simulator = self.end_use_load_profiles.simulate_building(
                ideal_loads=True,
                simulation_id=_simulation_id,
                output_directory=_output_directory,
                **_ideal_kwargs
            ).simulator

            # partial loads
            partial_loads_simulators: List[EndUseLoadProfilesEnergyPlusPartialLoadSimulator] = []
            kwargs_multiplier_minimum = partial_loads_kwargs.pop('multiplier_minimum', None)
            kwargs_multiplier_maximum = partial_loads_kwargs.pop('multiplier_maximum', None)
            kwargs_multiplier_probability = partial_loads_kwargs.pop('multiplier_probability', None)

            for i in range(partial_loads_simulations + 2):
                multiplier_minimum = kwargs_multiplier_minimum
                multiplier_maximum = kwargs_multiplier_maximum
                multiplier_probability = kwargs_multiplier_probability

                if i == 0:
                    reference_name = 'ideal'
                    multiplier_probability = 0
                
                elif i == 1:
                    reference_name = 'free_float'
                    multiplier_minimum = 0
                    multiplier_maximum = 0

                else:
                    reference_name = 'undercool_overcool_underheat_overheat'

                _simulation_id = f'{simulation_id}-partial-{reference_name}-{i}'
                _output_directory = os.path.join(output_directory, _simulation_id)
                _ = partial_loads_kwargs.pop('simulation_id', None)
                _ = partial_loads_kwargs.pop('output_directory', None)
                _ = partial_loads_kwargs.pop('random_seed', None)
                partial_loads_simulators.append(EndUseLoadProfilesEnergyPlusPartialLoadSimulator(
                    ideal_loads_simulator=deepcopy(ideal_simulator),
                    multiplier_minimum=multiplier_minimum,
                    multiplier_maximum=multiplier_maximum,
                    multiplier_probability=multiplier_probability,
                    simulation_id=_simulation_id,
                    output_directory=_output_directory,
                    random_seed=int(bldg_id) + i,
                    **partial_loads_kwargs
                ))
            
            EndUseLoadProfilesEnergyPlusPartialLoadSimulator.multi_simulate(partial_loads_simulators, max_workers=self.max_workers)

            simulators.append({
                'mechanical': mechanical_simulator,
                'ideal': ideal_simulator,
                'partial': partial_loads_simulators
            })
        
        return simulators

    def sample_buildings(
        self, filters: Mapping[str, List[Any]] = None, sample_method: SampleMethod = None, sample_count: int = None, duplicate_to_count: bool = None, 
        single_county: bool = None, single_family_detached: bool = None, **kwargs
    ) -> Tuple[List[int], List[int], Mapping[str, Any]]:
        sample_method = SampleMethod.RANDOM if sample_method is None else sample_method
        valid_sample_method = [v for v in SampleMethod]
        assert sample_method in valid_sample_method, f'Valid sample methods are {valid_sample_method}.'
        sample_count = 100 if sample_count is None else sample_count
        duplicate_to_count = True if duplicate_to_count is None else duplicate_to_count
        single_county = True if single_county is None else single_county
        single_family_detached = True if single_family_detached is None else single_family_detached
        
        nprs = np.random.RandomState(self.random_seed)
        sample_metadata = None

        if single_family_detached:
            filters = {} if filters is None else filters
            filters[self.__BUILDING_TYPE_COLUMN] =  [self.__SINGLE_FAMILY_BUILDING_TYPE_NAME]
        
        else:
            pass
        
        metadata = self.end_use_load_profiles.metadata.metadata.get(filters)

        if single_county:
            county = nprs.choice(metadata[self.__COUNTY_COLUMN].unique(), 1)[0]
            metadata = metadata[metadata[self.__COUNTY_COLUMN]==county].copy()

        else:
            pass

        if sample_method == SampleMethod.RANDOM:
            metadata['label'] = 0

        elif sample_method == SampleMethod.METADATA_CLUSTER_FREQUENCY:
            mc = MetadataClustering(self.end_use_load_profiles, metadata.index.tolist(), random_seed=self.random_seed, **kwargs)
            optimal_clusters, scores, labels = mc.cluster()
            sample_metadata = {
                'optimal_clusters': optimal_clusters,
                'scores': scores,
                'labels': labels
            }

            metadata = metadata.merge(labels[labels['clusters']==optimal_clusters].set_index('bldg_id'), left_index=True, right_index=True)

        metadata['label_count'] = metadata.groupby('label')['label'].transform('count')
        metadata = metadata.sample(sample_count, weights='label_count', replace=duplicate_to_count, random_state=self.random_seed)
        bldg_ids = metadata.index.tolist()
        labels = metadata['label'].tolist()

        return bldg_ids, labels, sample_metadata


"""Integration test for the charging constraints demo dataset."""

from pathlib import Path

import numpy as np
import pytest

pytest.importorskip("gymnasium")

from citylearn.citylearn import CityLearnEnv


DATASET_PATH = Path(__file__).resolve().parents[1] / "data/datasets/citylearn_three_phase_electrical_service_demo/schema.json"


def _find_action_index(env, charger_id: str) -> int:
    names = env.action_names[0]
    return names.index(f'electric_vehicle_storage_{charger_id}')


def test_charging_constraints_demo_dataset_runs():
    env = CityLearnEnv(str(DATASET_PATH), central_agent=True, episode_time_steps=4, random_seed=1)

    try:
        env.reset()
        charger_ids = ['charger_15_1', 'charger_15_2']
        actions = np.zeros(len(env.action_names[0]), dtype='float32')
        for cid in charger_ids:
            actions[_find_action_index(env, cid)] = 1.0

        reward = env.step([actions])[1]
        assert reward is not None

        building = next(b for b in env.buildings if b.name == 'Building_15')
        obs = building.observations(include_all=True, normalize=False, periodic_normalization=False)

        assert 'charging_constraint_violation_kwh' in obs
        assert obs['charging_constraint_violation_kwh'] >= 0.0
        one_hot_keys = [k for k in obs if k.startswith('charging_phase_one_hot_charger_15_1_')]
        assert one_hot_keys
        assert abs(sum(obs[k] for k in one_hot_keys) - 1.0) < 1e-6

    finally:
        env.close()


if __name__ == '__main__':
    test_charging_constraints_demo_dataset_runs()
    print('Charging constraints demo dataset test completed successfully.')



import logging
from typing import Any, List, Mapping
from gymnasium import spaces
import numpy as np
from citylearn.base import Environment
from citylearn.citylearn import CityLearnEnv

LOGGER = logging.getLogger()

class Agent(Environment):
    r"""Base agent class.

    Parameters
    ----------
    env : CityLearnEnv
        CityLearn environment.

    Other Parameters
    ----------------
    **kwargs : dict
        Other keyword arguments used to initialize super class.
    """
    
    def __init__(self, env: CityLearnEnv, **kwargs: Any):
        self.env = env
        self.observation_names = self.env.observation_names
        self.action_names = self.env.unwrapped.action_names
        self.observation_space = self.env.observation_space
        self.action_space = self.env.action_space
        self.episode_time_steps = self.env.unwrapped.time_steps
        self.building_metadata = self.env.unwrapped.get_metadata()['buildings']
        super().__init__(
            seconds_per_time_step=self.env.unwrapped.seconds_per_time_step,
            random_seed=self.env.unwrapped.random_seed,
            episode_tracker=self.env.unwrapped.episode_tracker,
            time_step_ratio=self.env.unwrapped.time_step_ratio
        )
        self.reset()

    @property
    def env(self) -> CityLearnEnv:
        """CityLearn environment."""

        return self.__env

    @property
    def observation_names(self) -> List[List[str]]:
        """Names of active observations that can be used to map observation values."""

        return self.__observation_names
    
    @property
    def action_names(self) -> List[List[str]]:
        """Names of active actions that can be used to map action values."""

        return self.__action_names

    @property
    def observation_space(self) -> List[spaces.Box]:
        """Format of valid observations."""

        return self.__observation_space

    @property
    def action_space(self) -> List[spaces.Box]:
        """Format of valid actions."""

        return self.__action_space
    
    @property
    def episode_time_steps(self) -> int:
        return self.__episode_time_steps

    @property
    def building_metadata(self) -> List[Mapping[str, Any]]:
        """Building(s) metadata."""

        return self.__building_metadata

    @property
    def action_dimension(self) -> List[int]:
        """Number of returned actions."""

        return [s.shape[0] for s in self.action_space]

    @property
    def actions(self) -> List[List[List[Any]]]:
        """Action history/time series."""

        return self.__actions
    
    @env.setter
    def env(self, env: CityLearnEnv):
        self.__env = env

    @observation_names.setter
    def observation_names(self, observation_names: List[List[str]]):
        self.__observation_names = observation_names

    @action_names.setter
    def action_names(self, action_names: List[List[str]]):
        self.__action_names = action_names

    @observation_space.setter
    def observation_space(self, observation_space: List[spaces.Box]):
        self.__observation_space = observation_space

    @action_space.setter
    def action_space(self, action_space: List[spaces.Box]):
        self.__action_space = action_space

    @episode_time_steps.setter
    def episode_time_steps(self, episode_time_steps: int):
        """Number of time steps in one episode."""

        self.__episode_time_steps = episode_time_steps

    @building_metadata.setter
    def building_metadata(self, building_metadata: List[Mapping[str, Any]]):
        self.__building_metadata = building_metadata

    @actions.setter
    def actions(self, actions: List[List[Any]]):
        for i in range(len(self.action_space)):
            self.__actions[i][self.time_step] = actions[i]

    def learn(self, episodes: int = None, deterministic: bool = None, deterministic_finish: bool = None, logging_level: int = None):
        """Train agent.

        Parameters
        ----------
        episodes: int, default: 1
            Number of training episode >= 1.
        deterministic: bool, default: False
            Indicator to take deterministic actions i.e. strictly exploit the learned policy.
        deterministic_finish: bool, default: False
            Indicator to take deterministic actions in the final episode.
        logging_level: int, default: 30
            Logging level where increasing the number silences lower level information.
        """
        
        episodes = 1 if episodes is None else episodes
        deterministic_finish = False if deterministic_finish is None else deterministic_finish
        deterministic = False if deterministic is None else deterministic
        self.__set_logger(logging_level)

        for episode in range(episodes):
            deterministic = deterministic or (deterministic_finish and episode >= episodes - 1)
            observations, _ = self.env.reset()
            self.episode_time_steps = self.episode_tracker.episode_time_steps
            terminated = self.env.terminated
            truncated = self.env.truncated
            time_step = 0
            rewards_list = []

            while not (terminated or truncated):
                actions = self.predict(observations, deterministic=deterministic)

                # apply actions to citylearn_env
                next_observations, rewards, terminated, truncated, _ = self.env.step(actions)
                rewards_list.append(rewards)

                # update
                if not deterministic:
                    self.update(observations, actions, rewards, next_observations, terminated=terminated, truncated=truncated)
                else:
                    pass

                observations = [o for o in next_observations]

                logging.debug(
                    f'Time step: {time_step + 1}/{self.episode_time_steps},'\
                        f' Episode: {episode + 1}/{episodes},'\
                            f' Actions: {actions},'\
                                f' Rewards: {rewards}'
                )

                time_step += 1

            if len(rewards_list) > 0:
                rewards = np.array(rewards_list, dtype='float')
                rewards_summary = {
                    'min': rewards.min(axis=0),
                    'max': rewards.max(axis=0),
                    'sum': rewards.sum(axis=0),
                    'mean': rewards.mean(axis=0)
                }
            else:
                reward_length = len(self.action_space)
                empty = np.zeros(reward_length, dtype='float')
                rewards_summary = {
                    'min': empty,
                    'max': empty,
                    'sum': empty,
                    'mean': empty,
                }
            logging.info(f'Completed episode: {episode + 1}/{episodes}, Reward: {rewards_summary}')

    def predict(self, observations: List[List[float]], deterministic: bool = None) -> List[List[float]]:
        """Provide actions for current time step.

        Return randomly sampled actions from `action_space`.
        
        Parameters
        ----------
        observations: List[List[float]]
            Environment observations
        deterministic: bool, default: False
            Wether to return purely exploitatative deterministic actions.

        Returns
        -------
        actions: List[List[float]]
            Action values
        """
        
        actions = [list(s.sample()) for s in self.action_space]
        self.actions = actions
        self.next_time_step()
        return actions
    
    def __set_logger(self, logging_level: int = None):
        """Set logging level."""

        logging_level = 30 if logging_level is None else logging_level
        assert logging_level >= 0, 'logging_level must be >= 0'
        LOGGER.setLevel(logging_level)

    def update(self, *args, **kwargs):
        """Update replay buffer and networks.
        
        Notes
        -----
        This implementation does nothing but is kept to keep the API for all agents similar during simulation.
        """

        pass

    def next_time_step(self):
        super().next_time_step()

        for i in range(len(self.action_space)):
            self.__actions[i].append([])

    def reset(self):
        super().reset()
        self.__actions = [[[]] for _ in self.action_space]

class BaselineAgent(Agent):
    r"""Agent class for business-as-usual simulation where the storage systems and heat pumps are not controlled.

    This agent will provide results for when there is no storage for load shifting and no heat pump partial load. 
    The storage actions prescribed will be 0.0 and the heat pump will have no action, i.e. `None`, causing it to 
    deliver the ideal load in the building time series files. 
    
    To ensure that the environment does not expect non-zero and non-null actions, the buildings in the parsed `env` 
    will be set to have no active actions. This means that you must initialize a new `env` if you want to simulate
    with a new agent type. 
    
    This agent class is best used to establish a baseline simulation that can then be compared 
    to RBC, RLC, or MPC control algorithms.

    Parameters
    ----------
    env : CityLearnEnv
        CityLearn environment.

    Other Parameters
    ----------------
    **kwargs : dict
        Other keyword arguments used to initialize super class.
    """

    def __init__(self, env: CityLearnEnv, **kwargs: Any):
        super().__init__(env, **kwargs)

    @Agent.env.setter
    def env(self, env: CityLearnEnv):
        Agent.env.fset(self, self.__deactivate_actions(env))

    def __deactivate_actions(self, env: CityLearnEnv) -> CityLearnEnv:
        for b in env.unwrapped.buildings:
            for a in b.action_metadata:
                b.action_metadata[a] = False

            b.action_space = b.estimate_action_space()
            b.observation_space = b.estimate_observation_space()

        return env

    def predict(self, observations: List[List[float]], deterministic: bool = None) -> List[List[float]]:
        actions = [[] for _ in self.action_names]
        self.actions = actions
        self.next_time_step()
        
        return actions



from pathlib import Path

import numpy as np
import pytest

pytest.importorskip("gymnasium")

from citylearn.citylearn import CityLearnEnv


SCHEMA = Path(__file__).resolve().parents[1] / "data/datasets/citylearn_challenge_2022_phase_all_plus_evs/schema.json"


def _zero_actions(env: CityLearnEnv):
    return [np.zeros(space.shape, dtype="float32") for space in env.action_space]


def test_agent_observations_use_lagged_endogenous_values():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        observations, _ = env.reset()
        names = env.observation_names[0]
        net_ix = names.index("net_electricity_consumption")
        soc_ix = names.index("electrical_storage_soc")

        assert observations[0][net_ix] == pytest.approx(float(env.buildings[0].net_electricity_consumption[0]))
        assert observations[0][soc_ix] == pytest.approx(float(env.buildings[0].electrical_storage.soc[0]))

        next_observations, _, _, _, _ = env.step(_zero_actions(env))
        assert env.time_step == 1
        assert next_observations[0][net_ix] == pytest.approx(float(env.buildings[0].net_electricity_consumption[0]))
        assert next_observations[0][soc_ix] == pytest.approx(float(env.buildings[0].electrical_storage.soc[0]))
    finally:
        env.close()


def test_include_all_observations_remain_on_current_transition_time():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        env.step(_zero_actions(env))
        building = env.buildings[0]
        t = env.time_step
        lagged_t = max(t - 1, 0)

        all_obs = building.observations(include_all=True, normalize=False, periodic_normalization=False)
        agent_obs = building.observations(include_all=False, normalize=False, periodic_normalization=False)

        assert all_obs["net_electricity_consumption"] == pytest.approx(float(building.net_electricity_consumption[t]))
        assert agent_obs["net_electricity_consumption"] == pytest.approx(float(building.net_electricity_consumption[lagged_t]))
        assert all_obs["electrical_storage_soc"] == pytest.approx(float(building.electrical_storage.soc[t]))
        assert agent_obs["electrical_storage_soc"] == pytest.approx(float(building.electrical_storage.soc[lagged_t]))
    finally:
        env.close()


def test_step_after_terminal_raises_runtime_error():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=2, random_seed=0)

    try:
        env.reset()
        _, _, terminated, truncated, _ = env.step(_zero_actions(env))
        assert terminated
        assert not truncated

        with pytest.raises(RuntimeError, match="reset"):
            env.step(_zero_actions(env))
    finally:
        env.close()


def test_step_raises_for_invalid_central_action_count():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        expected = env.action_space[0].shape[0]
        invalid = np.zeros(expected + 1, dtype="float32")

        with pytest.raises(AssertionError, match="Expected"):
            env.step([invalid])
    finally:
        env.close()


def test_step_accepts_central_numpy_action_vector():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        action = np.zeros(env.action_space[0].shape[0], dtype="float32")
        _, _, terminated, truncated, _ = env.step(action)
        assert not terminated
        assert not truncated
    finally:
        env.close()


def test_step_accepts_central_single_row_numpy_action_vector():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        action = np.zeros((1, env.action_space[0].shape[0]), dtype="float32")
        _, _, terminated, truncated, _ = env.step(action)
        assert not terminated
        assert not truncated
    finally:
        env.close()


def test_step_raises_for_missing_decentralized_action_vector():
    env = CityLearnEnv(str(SCHEMA), central_agent=False, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        actions = _zero_actions(env)

        if len(actions) < 2:
            pytest.skip("Dataset does not expose multiple buildings for decentralized action validation.")

        with pytest.raises(AssertionError, match="building action vectors"):
            env.step(actions[:-1])
    finally:
        env.close()


def test_step_accepts_valid_decentralized_action_vectors():
    env = CityLearnEnv(str(SCHEMA), central_agent=False, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        actions = _zero_actions(env)
        _, _, terminated, truncated, _ = env.step(actions)
        assert not terminated
        assert not truncated
    finally:
        env.close()



import torch
import torch.nn as nn

class LSTM(nn.Module):
    def __init__(self, n_features: int, n_output: int, drop_prob: float, seq_len: int, num_hidden: int, num_layers: int, weight_decay: float):
        super(LSTM, self).__init__()

        self.n_features = n_features  # number of inputs variable
        self.n_output = n_output  # number of output
        self.seq_len = seq_len  # lookback value
        self.n_hidden = num_hidden  # number of hidden states
        self.n_layers = num_layers  # number of LSTM layers (stacked)
        self.weight_decay = weight_decay

        self.l_lstm = nn.LSTM(input_size=self.n_features,
                              hidden_size=self.n_hidden,
                              num_layers=self.n_layers,
                              batch_first=True)
        self.dropout = torch.nn.Dropout(drop_prob)
        # LSTM Outputs: output, (h_n, c_n)
        # according to pytorch docs LSTM output is
        # (batch_size, seq_len, num_directions * hidden_size)
        # when considering batch_first = True
        self.l_linear = nn.Linear(self.n_hidden, n_output)

        # Add L2 regularization to the linear layer
        self.l_linear.weight_decay = self.weight_decay

    def init_hidden(self, batch_size, device):
        # even with batch_first = True this remains the same as docs
        hidden_state = torch.zeros(self.n_layers, batch_size, self.n_hidden)
        cell_state = torch.zeros(self.n_layers, batch_size, self.n_hidden)
        hidden_state = hidden_state.to(device)
        cell_state = cell_state.to(device)
        hidden = (hidden_state, cell_state)  # HIDDEN is defined as a TUPLE
        return hidden

    def forward(self, input_tensor, hidden_cell_tuple):
        batch_size, seq_len, _ = input_tensor.size()
        lstm_out, hidden_cell_tuple = self.l_lstm(input_tensor, hidden_cell_tuple)
        lstm_out = self.dropout(lstm_out)  # Applying dropout
        out = lstm_out[:, -1, :]  # I take only the last output vector, for each Batch
        out_linear = self.l_linear(out)
        return out_linear, hidden_cell_tuple


import argparse
import concurrent.futures
import datetime
import getpass
import importlib
import inspect
import logging
from multiprocessing import cpu_count
import os
from pathlib import Path
import shutil
import socket
import subprocess
import sys
from typing import Any, List, Mapping, Tuple, Union
import uuid
from citylearn.agents.base import Agent as CityLearnAgent
from citylearn.citylearn import CityLearnEnv
from citylearn.data import DataSet, get_settings
from citylearn.__init__ import __version__
from citylearn.utilities import FileHandler
import pandas as pd
import simplejson as json

try:
    from stable_baselines3.common.base_class import BaseAlgorithm as StableBaselines3Agent

except (ImportError, ModuleNotFoundError):
    pass

def run_work_order(work_order_filepath, max_workers=None, start_index=None, end_index=None, virtual_environment_path=None, windows_system=None):
    work_order_filepath = Path(work_order_filepath)
    
    if virtual_environment_path is not None:    
        if windows_system:
            virtual_environment_command = f'"{os.path.join(virtual_environment_path, "Scripts", "Activate.ps1")}"'
        
        else:
            virtual_environment_command = f'source "{os.path.join(virtual_environment_path, "bin", "activate")}"'
    
    else:
        virtual_environment_command = 'echo "No virtual environment"'

    with open(work_order_filepath,mode='r') as f:
        args = f.read()
    
    args = args.strip('\n').split('\n')
    start_index = 0 if start_index is None else start_index
    end_index = len(args) - 1 if end_index is None else end_index
    assert start_index <= end_index, 'start_index must be <= end_index'
    assert start_index < len(args), 'start_index must be < number of jobs'
    args = args[start_index:end_index + 1]
    args = [a for a in args if not a.startswith('#')]
    args = [f'{virtual_environment_command} && {a}' for a in args]
    max_workers = cpu_count() if max_workers is None else max_workers
    
    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
        logging.debug(f'Will use {max_workers} workers for job.')
        logging.debug(f'Pooling {len(args)} jobs to run in parallel...')
        results = [executor.submit(subprocess.run,**{'args':a, 'shell':True}) for a in args]
            
        for future in concurrent.futures.as_completed(results):
            try:
                logging.debug(future.result())
            
            except Exception as e:
                logging.debug(e)

class Simulator:
    def __init__(self, schema: str, agent_name: str = None, env_kwargs: Mapping[str, Any] = None, agent_kwargs: Mapping[str, Any] = None, wrappers: List[str] = None,
     time_series_variables: List[str] = None, simulation_id: str = None, output_directory: Union[Path, str] = None, agent_filepath: Union[Path, str] = None,
     random_seed: int = None, overwrite: bool = None
    ) -> None:
        self.schema = schema
        self.agent_name = agent_name
        self.env_kwargs = env_kwargs
        self.agent_kwargs = agent_kwargs
        self.random_seed = random_seed
        self.wrappers = wrappers
        self.time_series_variables = time_series_variables
        self.simulation_id = simulation_id
        self.overwrite = overwrite
        self.output_directory = output_directory
        self.agent_filepath = agent_filepath
        self.__reset()

    @property
    def schema(self) -> str:
        return self.__schema
    
    @property
    def agent_name(self) -> str:
        return self.__agent_name
    
    @property
    def env_kwargs(self) -> Mapping[str, Any]:
        return self.__env_kwargs
    
    @property
    def agent_kwargs(self) -> Mapping[str, Any]:
        return self.__agent_kwargs
    
    @property
    def wrappers(self) -> List[str]:
        return self.__wrappers
    
    @property
    def time_series_variables(self) -> List[str]:
        return self.__time_series_variables
    
    @property
    def simulation_id(self) -> str:
        return self.__simulation_id
    
    @property
    def output_directory(self) -> Path:
        return self.__output_directory
    
    @property
    def agent_filepath(self) -> Path:
        return self.__agent_filepath
    
    @property
    def random_seed(self) -> int:
        return self.__random_seed
    
    @property
    def overwrite(self) -> bool:
        return self.__overwrite
    
    @schema.setter
    def schema(self, value: str):
        self.__schema = value

    @agent_name.setter
    def agent_name(self, value: str):
        self.__agent_name = value

    @env_kwargs.setter
    def env_kwargs(self, value: Mapping[str, Any]):
        self.__env_kwargs = {} if value is None else value

    @agent_kwargs.setter
    def agent_kwargs(self, value: Mapping[str, Any]):
        self.__agent_kwargs = {} if value is None else value

    @wrappers.setter
    def wrappers(self, value:List[str]):
        self.__wrappers = [] if value is None else value

    @time_series_variables.setter
    def time_series_variables(self, value: List[str]):
        self.__time_series_variables = self.get_default_time_series_variables() if value is None else value

    @simulation_id.setter
    def simulation_id(self, value: str):
        self.__simulation_id = f'citylearn-simulation-{uuid.uuid4().hex}' if value is None else value

    @output_directory.setter
    def output_directory(self, value: Path):
        self.__output_directory = Path(os.path.join('citylearn_simulations', self.simulation_id))\
            if value is None else value
        
        if os.path.isdir(self.__output_directory) and self.overwrite:
            shutil.rmtree(self.__output_directory)
        
        else:
            pass

        os.makedirs(self.__output_directory, exist_ok=True)

    @agent_filepath.setter
    def agent_filepath(self, value: Path):
        self.__agent_filepath =  value if value is None else Path(value)

    @random_seed.setter
    def random_seed(self, value: int):
        self.__random_seed = value

        if self.random_seed is not None:
            self.env_kwargs['random_seed'] = self.random_seed

            if self.agent_name is not None:
                random_seed_name = 'seed' if 'stable_baselines3' in self.agent_name else 'random_seed'
                self.agent_kwargs[random_seed_name] = self.random_seed
            
            else:
                pass

        else:
            pass

    @overwrite.setter
    def overwrite(self, value: bool):
        self.__overwrite = True if value is None else value

    def __get_evaluation_summary(self):        
        return {
            'evaluation_episode_time_steps': [
                self.env.unwrapped.episode_tracker.episode_start_time_step, 
                self.env.unwrapped.episode_tracker.episode_end_time_step
            ],
            'evaluation_start_timestamp': self.__evaluation_start_timestamp,
            'evaluation_end_timestamp': self.__evaluation_end_timestamp,
            'evaluation': self.env.unwrapped.evaluate().pivot(index='name', columns='cost_function', values='value').to_dict('index'),
            'episode_reward_summary': self.env.unwrapped.episode_rewards[-1],
            'episode_rewards': self.env.unwrapped.rewards,
            'time_series': self.__get_time_series().to_dict('list'),
            'actions': self.__actions_list,
        }

    def __get_time_series(self) -> pd.DataFrame:
        data_list = []

        for b in self.env.unwrapped.buildings:
            data = {}

            for variable in self.time_series_variables:
                key = b
                values = variable.split('.')

                for i in range(len(values)):
                    if hasattr(key, values[i]):
                        value = getattr(key, values[i])
                        key = value
                    
                    else:
                        pass
                
                data[variable.replace('.', '_')] = value
            
            data = pd.DataFrame(data)
            data.insert(0, 'time_step', data.index)
            data.insert(1, 'building_name', b.name)
            data_list.append(data)

        return pd.concat(data_list, ignore_index=True)

    def __get_training_summary(self):
        return {
            'hostname': socket.gethostname(),
            'username': getpass.getuser(),
            'simulation_id': self.simulation_id,
            'agent': self.agent.__class__.__name__,
            'agent_kwargs': self.agent_kwargs,
            'wrappers': self.wrappers,
            'train_episodes': self.env.unwrapped.episode_tracker.episode,
            'train_episode_time_steps': self.env.unwrapped.episode_time_steps,
            'train_start_timestamp': self.__train_start_timestamp,
            'train_end_timestamp': self.__train_end_timestamp,
            'train_episode_reward_summary': self.env.unwrapped.episode_rewards,
            'env_metadata': self.env.unwrapped.get_metadata(),
        }

    def __evaluate(self):
        observations, _ = self.env.reset()
        actions_list = []
        self.__evaluation_start_timestamp = datetime.datetime.now(datetime.UTC)
        
        while not self.env.terminated:
            if isinstance(self.agent, CityLearnAgent):
                actions = self.agent.predict(observations, deterministic=True)
                actions_list.append(self.env.unwrapped._parse_actions(actions))
            
            elif isinstance(self.agent, StableBaselines3Agent):
                actions, _ = self.agent.predict(observations, deterministic=True)
                actions_list.append(self.env.unwrapped._parse_actions([actions]))

            else:
                raise Exception(f'Unknown agent type: {type(self.agent)}')

            observations, _, _, _, _ = self.env.step(actions)

        self.__evaluation_end_timestamp = datetime.datetime.now(datetime.UTC)
        self.__actions_list = actions_list

    def __train(self, episodes: int):
        kwargs = {}
        self.__train_start_timestamp = datetime.datetime.now(datetime.UTC)

        if isinstance(self.agent, CityLearnAgent):
            kwargs = {**kwargs, 'episodes': episodes}
            self.agent.learn(**kwargs)
        
        else:
            kwargs = {**kwargs, 'total_timesteps': episodes*self.env.unwrapped.time_steps}
            self.agent = self.agent.learn(**kwargs)
        
        self.__train_end_timestamp = datetime.datetime.now(datetime.UTC)

    def __save_agent(self):
        if isinstance(self.agent, CityLearnAgent):
            filepath = os.path.join(self.output_directory, f'{self.simulation_id}-agent.pkl')
            FileHandler.write_pickle(filepath, self.agent)
        
        else:
            filepath = os.path.join(self.output_directory, f'{self.simulation_id}-agent')
            self.agent.save(filepath)

    def __set_agent(self) -> Union[CityLearnAgent, StableBaselines3Agent]:
        if self.agent_filepath is None:
            agent = self.env.unwrapped.load_agent(
                agent=self.agent_name, 
                env=self.env,
                **self.agent_kwargs,
            )

        else:
            if str(self.agent_filepath).endswith('.pkl'):
                agent = FileHandler.read_pickle(self.agent_filepath)
                agent.env = self.env

            else:
                agent = StableBaselines3Agent.load(self.agent_filepath, self.env)

        return agent
    
    def __set_env(self) -> CityLearnEnv:
        env = CityLearnEnv(self.schema, **self.env_kwargs)

        for wrapper in self.wrappers:
            wrapper_module = '.'.join(wrapper.split('.')[0:-1])
            wrapper_name = wrapper.split('.')[-1]
            wrapper_constructor = getattr(importlib.import_module(wrapper_module), wrapper_name)
            env = wrapper_constructor(env)

        return env
    
    def __reset(self):
        self.env = self.__set_env()
        self.agent = self.__set_agent()
        self.__train_start_timestamp = None
        self.__train_end_timestamp = None
        self.__evaluation_start_timestamp = None
        self.__evaluation_end_timestamp = None
        self.__actions_list = None
    
    @classmethod
    def evaluate(cls, evaluation_episode_time_steps: Tuple[int, int] = None, **kwargs):
        kwargs['env_kwargs'] = {} if kwargs.get('env_kwargs') is None else kwargs['env_kwargs']

        if evaluation_episode_time_steps is not None:
            kwargs['env_kwargs']['episode_time_steps'] = [evaluation_episode_time_steps]
        
        else:
            pass
            

        simulator = cls(**kwargs)
        simulator.__evaluate()
        filepath = os.path.join(simulator.output_directory, f'{simulator.simulation_id}-evaluation.json')
        FileHandler.write_json(filepath, simulator.__get_evaluation_summary())

    @classmethod
    def train(cls, episodes: int = None, evaluate: bool = None, evaluation_episode_time_steps: Tuple[int, int] = None, save_agent: bool = None, **kwargs):
        simulator = cls(**kwargs)
        episodes = 1 if episodes is None else episodes
        simulator.__train(episodes)
        filepath = os.path.join(simulator.output_directory, f'{simulator.simulation_id}-train.json')
        FileHandler.write_json(filepath, simulator.__get_training_summary())

        if save_agent:
            simulator.__save_agent()

        else:
            pass

        if evaluate:
            evaluation_episode_time_steps = [[
                simulator.env.unwrapped.episode_tracker.simulation_start_time_step, 
                simulator.env.unwrapped.episode_tracker.simulation_end_time_step
            ]] if evaluation_episode_time_steps is None else [evaluation_episode_time_steps]
            kwargs['env_kwargs'] = {} if kwargs.get('env_kwargs') is None else kwargs['env_kwargs']
            kwargs['overwrite'] = False
            kwargs['output_directory'] = simulator.output_directory
            kwargs['env_kwargs']['episode_time_steps'] = evaluation_episode_time_steps
            kwargs['simulation_id'] = f'{simulator.simulation_id}'
            simulator.evaluate(**kwargs)

        else:
            pass

    @staticmethod
    def get_default_time_series_variables():
        return get_settings()['default_time_series_variables']

def main():
    parser = argparse.ArgumentParser(
        prog='citylearn', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=(
            'An open source Farama Foundation Gymnasium environment for benchmarking distributed energy resource '
            'control algorithms to provide energy flexibility in a district of buildings. '
            'Compatible with training and evaluating internally defined CityLearn agents in `citylearn.agents`, '
            'user-defined agents that inherit from `citylearn.agents.base.Agent` and use the same interface as it, and agents '
            'provided by stable-baselines3.'))
    parser.add_argument('--version', action='version', version='%(prog)s' + f' {__version__}')
    subparsers = parser.add_subparsers(title='subcommands', required=True, dest='subcommands')

    # run many simulations in parallel
    subparser_run_work_order = subparsers.add_parser(
        'run_work_order', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, 
        help='Run commands in parallel. Useful for running many `citylearn simulate` commands in parallel.'
    )
    subparser_run_work_order.add_argument('work_order_filepath', type=Path, help=(
        'Filepath to script containing list of commands to be run in parallel with each command defined on a new line.'))
    subparser_run_work_order.add_argument('-w', '--max_workers', dest='max_workers', type=int, help=(
        'Maximum number of commands to run at a time. Default is the number of CPUs.'))
    subparser_run_work_order.add_argument('-is', '--start_index', default=0, dest='start_index', type=int, help=(
        'Line index of first command to execute. Commands above this index are not executed. '
        'The default is to execute from the first line.'))
    subparser_run_work_order.add_argument('-ie', '--end_index', dest='end_index', type=int, help=(
        'Line index of last command to execute. Commands below this index are not exectued. '
        'The default is to execute till the last line.'))
    subparser_run_work_order.set_defaults(func=run_work_order)

    # get names of datasets
    subparser_datasets = subparsers.add_parser(
        'list_datasets', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, 
        help='Lists available dataset names that can be parsed as `schema` in `citylearn simulate schema`.'
    )
    subparser_datasets.set_defaults(func=DataSet().get_dataset_names)

     # get default time series variables
    subparser_time_series_variables = subparsers.add_parser(
        'list_default_time_series_variables', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, 
        help='Lists the default time series variables that will be reported and saved in a `JSON` file post-evaluation.'
    )
    subparser_time_series_variables.set_defaults(func=Simulator.get_default_time_series_variables)

    # run one simulation
    subparser_simulate = subparsers.add_parser(
        'simulate', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, 
        help='Train or evaluate a trained agent against an environment.'
    )
    subparser_simulate.add_argument('schema', type=str, help=(
        'Name of CityLearn dataset or filepath to a schema. Call `citylearn list_datasets` to get list of valid dataset names.'))
    subparser_simulate.add_argument('-a', '--agent_name', dest='agent_name', default='citylearn.agents.base.BaselineAgent', type=str, help=(
        'Name path to agent. Currently only compatible with internally defined CityLearn agents in `citylearn.agents`, '
        'user-defined agents that inherit from `citylearn.agents.base.Agent` and use the same interface as it, and agents '
        'provided by stable-baselines3. To use stable-baselines3 agents, make sure to run `pip install stable-baselines3` '
        'before using the `simulate command.`'))
    subparser_simulate.add_argument('-ke', '--env_kwargs', dest='env_kwargs', type=json.loads, help=(
        'Initialization parameters for`citylearn.citylearn.CityLearnEnv`.'))
    subparser_simulate.add_argument('-ka', '--agent_kwargs', dest='agent_kwargs', type=json.loads, help=(
        'Initialization parameters for agent class.'))
    subparser_simulate.add_argument('-w', '--wrappers', dest='wrappers', type=str, nargs='+', help=(
        'Name path to environment wrappers e.g., \'citylearn.wrappers.ClippedObservationWrapper\'.'))
    subparser_simulate.add_argument('-tv', '--time_series_variables', dest='time_series_variables', type=str, nargs='+', help=(
        'Names of building-level time series properties to be stored in the evaluation `JSON` post-evaluation. '
        'Call `citylearn list_default_time_series_variables` to see the default variable in use.'))
    subparser_simulate.add_argument('-sid', '--simulation_id', dest='simulation_id', type=str, help=(
        'SImulation reference ID used in directory and file names.' ))
    subparser_simulate.add_argument('-fa', '--agent_filepath', dest='agent_filepath', type=str, help=(
        'Filepath to previously saved agent to use for training or evaluation.'))
    subparser_simulate.add_argument('-d', '--output_directory', dest='output_directory', type=str, help=(
        'Directory to save all simulation output to.'))
    subparser_simulate.add_argument('-te', '--evaluation_episode_time_steps', dest='evaluation_episode_time_steps', type=int, nargs=2,
        action='append', help=('Start and end time steps in data set to evaluate on otherwise, the agent is evaluated on entire dataset.'))
    subparser_simulate.add_argument('-p', '--append', dest='overwrite', action='store_false', help=(
        'Add to output for existing simulation with `simulation_id` i.e. do not overwrite.'))
    subparser_simulate.add_argument('-rs', '--random_seed', dest='random_seed', type=int, help=(
        'Random seed used during environment and agent initialization.'))
    simulation_subparsers = subparser_simulate.add_subparsers(title='simulate subcommands', required=True, dest='subcommands')

    # -> train an agent
    subparser_train = simulation_subparsers.add_parser(
        'train', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, 
        help='Train an agent.'
    )
    subparser_train.add_argument('-e', '--episodes', dest='episodes', type=int, help='Number of training episodes/epochs.')
    subparser_train.add_argument('--save_agent', dest='save_agent', action='store_true', help='Whether to save agent to disk at the end of training.')
    subparser_train.add_argument('--evaluate', dest='evaluate', action='store_true', help=(
        'Whether to run deterministic evaluation for one episode at the end of training.'))
    subparser_train.set_defaults(func=Simulator.train)

    # -> evaluate a trained agent
    subparser_evaluate = simulation_subparsers.add_parser(
        'evaluate', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, 
        help='Deterministically evaluate an agent.'
    )
    subparser_evaluate.set_defaults(func=Simulator.evaluate)

    argv = sys.argv
    argv = ['sphinx-build' in a for a in argv]
    sphinx = any(argv)

    if not sphinx:
        args = parser.parse_args()
        arg_spec = inspect.getfullargspec(args.func)
        kwargs = {key:value for (key, value) in args._get_kwargs() 
            if (key in arg_spec.args or (arg_spec.varkw is not None and key not in ['func','subcommands']))
        }

        return args.func(**kwargs)
    
    else:
        return parser

if __name__ == '__main__':
    sys.exit(main())


from pathlib import Path

import numpy as np
import pytest

pytest.importorskip("gymnasium")

from citylearn.citylearn import CityLearnEnv


SCHEMA = Path(__file__).resolve().parents[1] / "data/datasets/citylearn_challenge_2022_phase_all_plus_evs/schema.json"


def _assert_length_consistency(env: CityLearnEnv):
    t = env.time_step
    for building in env.buildings:
        assert len(building.net_electricity_consumption) == t + 1
        assert len(building.cooling_electricity_consumption) == t + 1
        assert len(building.heating_electricity_consumption) == t + 1
        assert len(building.dhw_electricity_consumption) == t + 1
        assert len(building.non_shiftable_load_electricity_consumption) == t + 1
        assert len(building.electrical_storage_electricity_consumption) == t + 1
        assert len(building.chargers_electricity_consumption) == t + 1
        assert len(building.washing_machines_electricity_consumption) == t + 1


def test_series_integrity_reset_and_step():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        assert env.time_step == 0
        assert len(env.net_electricity_consumption) == 1
        assert len(env.net_electricity_consumption_cost) == 1
        assert len(env.net_electricity_consumption_emission) == 1
        _assert_length_consistency(env)

        zeros = [np.zeros(env.action_space[0].shape[0], dtype="float32")]

        env.step(zeros)
        assert env.time_step == 1
        assert len(env.net_electricity_consumption) == 1
        assert len(env.net_electricity_consumption_cost) == 1
        assert len(env.net_electricity_consumption_emission) == 1
        _assert_length_consistency(env)

        env.step(zeros)
        assert env.time_step == 2
        assert len(env.net_electricity_consumption) == 2
        assert len(env.net_electricity_consumption_cost) == 2
        assert len(env.net_electricity_consumption_emission) == 2
        _assert_length_consistency(env)
    finally:
        env.close()


def test_bess_first_step_not_double_counted():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        action = np.zeros(env.action_space[0].shape[0], dtype="float32")

        offset = 0
        target_building = None
        target_action_index = None

        for building in env.buildings:
            action_count = len(building.active_actions)
            if "electrical_storage" in building.active_actions:
                local_index = building.active_actions.index("electrical_storage")
                target_action_index = offset + local_index
                target_building = building
                break
            offset += action_count

        assert target_building is not None
        assert target_action_index is not None
        action[target_action_index] = 0.5

        env.step([action])

        t = 0
        expected = target_building.electrical_storage.energy_balance[t]
        actual = target_building.electrical_storage_electricity_consumption[t]

        assert abs(expected) > 1e-9
        assert actual == pytest.approx(expected, abs=1e-6)
    finally:
        env.close()


def test_non_shiftable_first_step_not_double_counted():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        action = [np.zeros(env.action_space[0].shape[0], dtype="float32")]
        env.step(action)

        for building in env.buildings:
            expected = building.energy_to_non_shiftable_load[0]
            actual = building.non_shiftable_load_electricity_consumption[0]
            assert actual == pytest.approx(expected, abs=1e-6)
    finally:
        env.close()


def test_non_shiftable_t0_update_variables_idempotent():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        before = {
            building.name: float(building.non_shiftable_load_electricity_consumption[0])
            for building in env.buildings
        }

        env.update_variables()

        after = {
            building.name: float(building.non_shiftable_load_electricity_consumption[0])
            for building in env.buildings
        }
        assert after == pytest.approx(before, abs=1e-6)
    finally:
        env.close()


def test_terminal_series_exclude_uncommitted_tail_slot():
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=6, random_seed=0)

    try:
        env.reset()
        zeros = [np.zeros(env.action_space[0].shape[0], dtype="float32")]

        while not env.terminated:
            _, _, terminated, truncated, _ = env.step(zeros)
            if terminated or truncated:
                break

        assert env.time_step == env.time_steps - 1

        for building in env.buildings:
            assert building.time_step == env.time_step - 1
            assert len(building.net_electricity_consumption) == env.time_step
            assert len(building.net_electricity_consumption_cost) == env.time_step
            assert len(building.net_electricity_consumption_emission) == env.time_step
            assert len(building.electrical_storage_electricity_consumption) == env.time_step
    finally:
        env.close()



import json
import shutil
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

pytest.importorskip("gymnasium")

from citylearn.citylearn import CityLearnEnv


MINUTE_DATASET_DIR = Path(__file__).resolve().parents[1] / "tests" / "data" / "minute_ev_demo"


def _clone_minute_schema(tmp_path: Path, name: str, mutator=None) -> Path:
    dataset_dir = tmp_path / name
    shutil.copytree(MINUTE_DATASET_DIR, dataset_dir)
    schema_path = dataset_dir / "schema.json"

    with open(schema_path, "r", encoding="utf-8") as f:
        schema = json.load(f)

    if mutator is not None:
        mutator(schema)

    with open(schema_path, "w", encoding="utf-8") as f:
        json.dump(schema, f, indent=2)

    return schema_path


def _rollout_zero_actions(env: CityLearnEnv):
    env.reset()

    while not env.terminated:
        action = np.zeros(len(env.action_names[0]), dtype="float32")
        env.step([action])


def _build_two_building_market_schema(tmp_path: Path) -> Path:
    dataset_dir = tmp_path / "market_dataset"
    dataset_dir.mkdir(parents=True, exist_ok=True)

    weather = pd.read_csv(MINUTE_DATASET_DIR / "weather.csv").iloc[:2].copy()
    weather.to_csv(dataset_dir / "weather.csv", index=False)

    carbon = pd.read_csv(MINUTE_DATASET_DIR / "carbon_intensity.csv").iloc[:2].copy()
    carbon.to_csv(dataset_dir / "carbon_intensity.csv", index=False)

    pricing = pd.DataFrame(
        {
            "electricity_pricing": [0.5, 0.5],
            "electricity_pricing_predicted_1": [0.5, 0.5],
            "electricity_pricing_predicted_2": [0.5, 0.5],
            "electricity_pricing_predicted_3": [0.5, 0.5],
        }
    )
    pricing.to_csv(dataset_dir / "pricing.csv", index=False)

    building_a = pd.DataFrame(
        {
            "month": [1, 1],
            "hour": [0, 1],
            "minutes": [0, 0],
            "day_type": [1, 1],
            "daylight_savings_status": [0, 0],
            "indoor_dry_bulb_temperature": [21.0, 21.0],
            "average_unmet_cooling_setpoint_difference": [0.0, 0.0],
            "indoor_relative_humidity": [45.0, 45.0],
            "non_shiftable_load": [2.0, 2.0],
            "dhw_demand": [0.0, 0.0],
            "cooling_demand": [0.0, 0.0],
            "heating_demand": [0.0, 0.0],
            "solar_generation": [0.0, 0.0],
        }
    )
    building_b = building_a.copy()
    building_b["non_shiftable_load"] = [0.0, 0.0]
    building_b["solar_generation"] = [2000.0, 2000.0]

    building_a.to_csv(dataset_dir / "Building_A.csv", index=False)
    building_b.to_csv(dataset_dir / "Building_B.csv", index=False)

    schema = {
        "random_seed": 0,
        "root_directory": None,
        "central_agent": True,
        "simulation_start_time_step": 0,
        "simulation_end_time_step": 1,
        "episode_time_steps": 2,
        "rolling_episode_split": False,
        "random_episode_split": False,
        "seconds_per_time_step": 3600,
        "observations": {
            "month": {"active": True, "shared_in_central_agent": True},
            "hour": {"active": True, "shared_in_central_agent": True},
            "minutes": {"active": True, "shared_in_central_agent": True},
            "day_type": {"active": True, "shared_in_central_agent": True},
            "outdoor_dry_bulb_temperature": {"active": True, "shared_in_central_agent": True},
            "non_shiftable_load": {"active": True, "shared_in_central_agent": False},
            "solar_generation": {"active": True, "shared_in_central_agent": False},
            "net_electricity_consumption": {"active": True, "shared_in_central_agent": False},
            "electricity_pricing": {"active": True, "shared_in_central_agent": True},
        },
        "actions": {
            "electrical_storage": {"active": False},
        },
        "reward_function": {
            "type": "citylearn.reward_function.RewardFunction",
            "attributes": {},
        },
        "community_market": {
            "enabled": True,
            "intra_community_sell_ratio": 0.8,
            "grid_export_price": 0.0,
        },
        "buildings": {
            "Building_A": {
                "include": True,
                "energy_simulation": "Building_A.csv",
                "weather": "weather.csv",
                "carbon_intensity": "carbon_intensity.csv",
                "pricing": "pricing.csv",
                "inactive_observations": [],
                "inactive_actions": [],
                "pv": {
                    "type": "citylearn.energy_model.PV",
                    "autosize": False,
                    "attributes": {"nominal_power": 0.0},
                },
            },
            "Building_B": {
                "include": True,
                "energy_simulation": "Building_B.csv",
                "weather": "weather.csv",
                "carbon_intensity": "carbon_intensity.csv",
                "pricing": "pricing.csv",
                "inactive_observations": [],
                "inactive_actions": [],
                "pv": {
                    "type": "citylearn.energy_model.PV",
                    "autosize": False,
                    "attributes": {"nominal_power": 1.0},
                },
            },
        },
    }

    schema_path = dataset_dir / "schema.json"
    with open(schema_path, "w", encoding="utf-8") as f:
        json.dump(schema, f, indent=2)

    return schema_path


def test_disabled_community_market_keeps_legacy_costs(tmp_path: Path):
    baseline_schema = _clone_minute_schema(tmp_path, "baseline")
    disabled_schema = _clone_minute_schema(
        tmp_path,
        "disabled_market",
        mutator=lambda s: s.update(
            {
                "community_market": {
                    "enabled": False,
                    "intra_community_sell_ratio": 0.2,
                    "grid_export_price": 0.9,
                }
            }
        ),
    )

    env_baseline = CityLearnEnv(str(baseline_schema), central_agent=True, episode_time_steps=4, random_seed=0)
    env_disabled = CityLearnEnv(str(disabled_schema), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        _rollout_zero_actions(env_baseline)
        _rollout_zero_actions(env_disabled)

        np.testing.assert_allclose(
            np.asarray(env_baseline.net_electricity_consumption_cost, dtype="float64"),
            np.asarray(env_disabled.net_electricity_consumption_cost, dtype="float64"),
            rtol=1e-9,
            atol=1e-9,
        )
    finally:
        env_baseline.close()
        env_disabled.close()


def test_disabled_community_market_with_string_false_keeps_legacy_costs(tmp_path: Path):
    baseline_schema = _clone_minute_schema(tmp_path, "baseline_string_false")
    disabled_schema = _clone_minute_schema(
        tmp_path,
        "disabled_market_string_false",
        mutator=lambda s: s.update(
            {
                "community_market": {
                    "enabled": "false",
                    "intra_community_sell_ratio": 0.2,
                    "grid_export_price": 0.9,
                }
            }
        ),
    )

    env_baseline = CityLearnEnv(str(baseline_schema), central_agent=True, episode_time_steps=4, random_seed=0)
    env_disabled = CityLearnEnv(str(disabled_schema), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        _rollout_zero_actions(env_baseline)
        _rollout_zero_actions(env_disabled)

        np.testing.assert_allclose(
            np.asarray(env_baseline.net_electricity_consumption_cost, dtype="float64"),
            np.asarray(env_disabled.net_electricity_consumption_cost, dtype="float64"),
            rtol=1e-9,
            atol=1e-9,
        )
    finally:
        env_baseline.close()
        env_disabled.close()


def test_single_phase_rejects_non_l1_assets(tmp_path: Path):
    schema_path = _clone_minute_schema(
        tmp_path,
        "single_phase_invalid",
        mutator=lambda s: (
            s["buildings"]["Building_1"].update(
                {
                    "electrical_service": {
                        "mode": "single_phase",
                        "limits": {"total": {"import_kw": 8.0, "export_kw": 8.0}},
                    }
                }
            ),
            s["buildings"]["Building_1"]["chargers"]["charger_1_1"]["attributes"].update({"phase_connection": "L2"}),
        ),
    )

    with pytest.raises(ValueError, match="single_phase"):
        CityLearnEnv(str(schema_path), central_agent=True, episode_time_steps=4, random_seed=0)


def test_three_phase_rejects_invalid_per_phase_keys(tmp_path: Path):
    schema_path = _clone_minute_schema(
        tmp_path,
        "three_phase_invalid_key",
        mutator=lambda s: s["buildings"]["Building_1"].update(
            {
                "electrical_service": {
                    "mode": "three_phase",
                    "limits": {
                        "total": {"import_kw": 8.0, "export_kw": 8.0},
                        "per_phase": {"L1": {"import_kw": 3.0}, "L4": {"import_kw": 3.0}},
                    },
                }
            }
        ),
    )

    with pytest.raises(ValueError, match="L1/L2/L3"):
        CityLearnEnv(str(schema_path), central_agent=True, episode_time_steps=4, random_seed=0)


def test_electrical_service_rejects_nan_limits(tmp_path: Path):
    def _mutate(schema):
        building = schema["buildings"]["Building_1"]
        building["electrical_service"] = {
            "mode": "three_phase",
            "limits": {
                "total": {"import_kw": "NaN", "export_kw": 10.0},
                "per_phase": {},
            },
        }

    schema_path = _clone_minute_schema(tmp_path, "electrical_service_nan_limit", mutator=_mutate)

    with pytest.raises(ValueError, match="cannot be NaN"):
        CityLearnEnv(str(schema_path), central_agent=True, episode_time_steps=4, random_seed=0)


def test_electrical_service_positive_infinite_limits_are_treated_as_unbounded(tmp_path: Path):
    def _mutate(schema):
        building = schema["buildings"]["Building_1"]
        building["electrical_service"] = {
            "mode": "three_phase",
            "limits": {
                "total": {"import_kw": "inf", "export_kw": "inf"},
                "per_phase": {
                    "L1": {"import_kw": "inf", "export_kw": "inf"},
                    "L2": {"import_kw": "inf", "export_kw": "inf"},
                    "L3": {"import_kw": "inf", "export_kw": "inf"},
                },
            },
            "observations": {"headroom": True, "violation": True},
        }
        building["electrical_storage"]["attributes"]["phase_connection"] = "all_phases"

    schema_path = _clone_minute_schema(tmp_path, "electrical_service_infinite_limit", mutator=_mutate)
    env = CityLearnEnv(str(schema_path), central_agent=True, episode_time_steps=4, random_seed=0)

    try:
        env.reset()
        action_names = env.action_names[0]
        actions = np.zeros(len(action_names), dtype="float32")
        actions[action_names.index("electrical_storage")] = 1.0
        ev_action_name = next(name for name in action_names if name.startswith("electric_vehicle_storage_"))
        actions[action_names.index(ev_action_name)] = 1.0
        env.step([actions])

        building = env.buildings[0]
        state = building._charging_constraints_state
        assert state is not None
        assert state["building_headroom_kw"] is None
        assert state["building_export_headroom_kw"] is None
        assert state["phase_headroom_kw"]["L1"] is None
        assert state["phase_headroom_kw"]["L2"] is None
        assert state["phase_headroom_kw"]["L3"] is None
        assert state["phase_export_headroom_kw"]["L1"] is None
        assert state["phase_export_headroom_kw"]["L2"] is None
        assert state["phase_export_headroom_kw"]["L3"] is None
        assert np.isfinite(state["total_power_kw"])
        assert np.isfinite(building._charging_constraint_last_penalty_kwh)
        assert building._charging_constraint_last_penalty_kwh == pytest.approx(0.0, abs=1e-6)
    finally:
        env.close()


def test_three_phase_limits_clip_controllable_actions(tmp_path: Path):
    def _mutate(schema):
        building = schema["buildings"]["Building_1"]
        building["electrical_service"] = {
            "mode": "three_phase",
            "default_split": "balanced",
            "limits": {
                "total": {"import_kw": 2.5, "export_kw": 2.5},
                "per_phase": {
                    "L1": {"import_kw": 1.0, "export_kw": 1.0},
                    "L2": {"import_kw": 1.0, "export_kw": 1.0},
                    "L3": {"import_kw": 1.0, "export_kw": 1.0},
                },
            },
            "observations": {"headroom": True, "violation": True},
        }
        building["chargers"]["charger_1_1"]["attributes"]["phase_connection"] = "L1"
        building["electrical_storage"]["attributes"]["phase_connection"] = "all_phases"

    schema_path = _clone_minute_schema(tmp_path, "three_phase_clip", mutator=_mutate)
    env = CityLearnEnv(str(schema_path), central_agent=True, episode_time_steps=4, random_seed=1)

    try:
        env.reset()
        action_names = env.action_names[0]
        actions = np.zeros(len(action_names), dtype="float32")
        actions[action_names.index("electrical_storage")] = 1.0
        ev_action_name = next(name for name in action_names if name.startswith("electric_vehicle_storage_"))
        actions[action_names.index(ev_action_name)] = 1.0
        env.step([actions])

        building = env.buildings[0]
        state = building._charging_constraints_state
        assert state is not None
        assert state["total_power_kw"] <= 2.5 + 1e-6
        assert state["phase_power_kw"]["L1"] <= 1.0 + 1e-6
        assert state["phase_power_kw"]["L2"] <= 1.0 + 1e-6
        assert state["phase_power_kw"]["L3"] <= 1.0 + 1e-6
        assert building._charging_constraint_last_penalty_kwh == pytest.approx(0.0, abs=1e-6)

        t = building.time_step - 1
        charger = building.electric_vehicle_chargers[0]
        commanded_kwh = charger.past_charging_action_values_kwh[t]
        assert commanded_kwh < (charger.max_charging_power * (building.seconds_per_time_step / 3600.0))
    finally:
        env.close()


def test_residual_violation_when_non_controllable_exceeds_limit(tmp_path: Path):
    def _mutate(schema):
        building = schema["buildings"]["Building_1"]
        building["electrical_service"] = {
            "mode": "three_phase",
            "default_split": "balanced",
            "limits": {
                "total": {"import_kw": 0.1, "export_kw": 2.0},
                "per_phase": {},
            },
            "observations": {"headroom": True, "violation": True},
        }
        building["electrical_storage"]["attributes"]["phase_connection"] = "all_phases"

    schema_path = _clone_minute_schema(tmp_path, "residual_violation", mutator=_mutate)
    env = CityLearnEnv(str(schema_path), central_agent=True, episode_time_steps=4, random_seed=2)

    try:
        env.reset()
        actions = np.zeros(len(env.action_names[0]), dtype="float32")
        env.step([actions])

        building = env.buildings[0]
        assert building._charging_constraint_last_penalty_kwh > 0.0
        state = building._charging_constraints_state
        assert state["total_power_kw"] > 0.1
        obs = building.observations(include_all=True, normalize=False, periodic_normalization=False)
        assert obs["charging_constraint_violation_kwh"] > 0.0
    finally:
        env.close()


def test_community_market_settlement_matches_expected_values(tmp_path: Path):
    schema_path = _build_two_building_market_schema(tmp_path)
    env = CityLearnEnv(str(schema_path), central_agent=True, episode_time_steps=2, random_seed=0)

    try:
        env.reset()
        env.step([np.zeros(len(env.action_names[0]), dtype="float32")])

        building_a = next(building for building in env.buildings if building.name == "Building_A")
        building_b = next(building for building in env.buildings if building.name == "Building_B")
        t = env.time_step - 1

        net_a = float(building_a.net_electricity_consumption[t])
        net_b = float(building_b.net_electricity_consumption[t])
        imports = np.array([max(net_a, 0.0), max(net_b, 0.0)], dtype="float64")
        exports = np.array([max(-net_a, 0.0), max(-net_b, 0.0)], dtype="float64")

        traded = min(float(imports.sum()), float(exports.sum()))
        local_import = imports * (traded / max(float(imports.sum()), 1e-12))
        local_export = exports * (traded / max(float(exports.sum()), 1e-12))
        grid_import_remaining = imports - local_import
        grid_export_remaining = exports - local_export

        p_grid = 0.5
        p_local = 0.8 * p_grid
        p_export = 0.0
        expected_cost_a = (
            grid_import_remaining[0] * p_grid
            + local_import[0] * p_local
            - local_export[0] * p_local
            - grid_export_remaining[0] * p_export
        )
        expected_cost_b = (
            grid_import_remaining[1] * p_grid
            + local_import[1] * p_local
            - local_export[1] * p_local
            - grid_export_remaining[1] * p_export
        )

        assert building_a.net_electricity_consumption_cost[t] == pytest.approx(expected_cost_a, abs=1e-6)
        assert building_b.net_electricity_consumption_cost[t] == pytest.approx(expected_cost_b, abs=1e-6)
        assert env.net_electricity_consumption_cost[t] == pytest.approx(expected_cost_a + expected_cost_b, abs=1e-6)
    finally:
        env.close()


def test_community_market_equal_share_allocator():
    from citylearn.internal.runtime import CityLearnRuntimeService

    allocations = CityLearnRuntimeService._allocate_equal_share_import(
        np.array([7.0, 7.0, 7.0, 7.0], dtype='float64'),
        20.0,
    )
    np.testing.assert_allclose(allocations, np.array([5.0, 5.0, 5.0, 5.0], dtype='float64'), atol=1e-9, rtol=1e-9)

    capped_allocations = CityLearnRuntimeService._allocate_equal_share_import(
        np.array([2.0, 8.0, 8.0, 8.0], dtype='float64'),
        20.0,
    )
    np.testing.assert_allclose(capped_allocations, np.array([2.0, 6.0, 6.0, 6.0], dtype='float64'), atol=1e-9, rtol=1e-9)



import random
from typing import Any, Mapping, Tuple
import numpy as np
import pandas as pd
import torch
from citylearn.end_use_load_profiles.lstm_model.model import LSTM
from citylearn.end_use_load_profiles.lstm_model.model_generation import run

def run_one_model(config: Mapping[str, Any], df: pd.DataFrame, seed: int) -> Mapping[str, Any]:
    model, observation_metadata, error_metrics = get_model(config, df, seed)
    
    return {
        'model': model,
        'attributes': {
            'hidden_size': config['hidden_size'],
            'num_layers': config['num_layer'],
            'lookback': config['lb'],
            'input_observation_names': observation_metadata['input_observation_names'],
            'input_normalization_minimum': observation_metadata['input_normalization_minimum'],
            'input_normalization_maximum': observation_metadata['input_normalization_maximum']
        },
        'error_metrics': error_metrics
    }

def get_model(config: Mapping[str, Any], df: pd.DataFrame, seed) -> Tuple[LSTM, Mapping[str, Any], Mapping[str, float]]:
    set_random_seeds(seed)
    lstm, observation_metadata, error = run(config, df)
    
    return lstm, observation_metadata, error

def set_random_seeds(seed: int, benchmark: bool = None, deterministic: bool = None):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False if benchmark is None else benchmark
    torch.backends.cudnn.deterministic = True if deterministic is None else deterministic


from __future__ import annotations

from dataclasses import dataclass
import hashlib
import importlib
import os
from typing import TYPE_CHECKING, Any, List, Mapping, Tuple, Union

import numpy as np
import pandas as pd

from citylearn.base import EpisodeTracker
from citylearn.building import Building
from citylearn.data import (
    CarbonIntensity,
    ChargerSimulation,
    DataSet,
    EnergySimulation,
    LogisticRegressionOccupantParameters,
    Pricing,
    WashingMachineSimulation,
    Weather,
)
from citylearn.electric_vehicle import ElectricVehicle
from citylearn.energy_model import Battery, PV, WashingMachine
from citylearn.reward_function import MultiBuildingRewardFunction, RewardFunction
from citylearn.utilities import parse_bool

if TYPE_CHECKING:
    from citylearn.citylearn import CityLearnEnv


@dataclass
class LoadContext:
    """Lightweight holder for loading inputs."""

    schema: Mapping[str, Any]
    kwargs: Mapping[str, Any]


class CityLearnLoadingService:
    """Internal loader service that builds env components from schema."""

    def __init__(self, env: "CityLearnEnv"):
        self.env = env

    def load(
        self,
        schema: Mapping[str, Any],
        **kwargs,
    ) -> Tuple[
        Union[os.PathLike, str],
        List[Building],
        List[ElectricVehicle],
        Union[int, List[Tuple[int, int]]],
        bool,
        bool,
        float,
        RewardFunction,
        bool,
        List[str],
        EpisodeTracker,
    ]:
        """Return env objects as defined by schema."""

        schema['root_directory'] = kwargs['root_directory'] if kwargs.get('root_directory') is not None else schema['root_directory']
        schema['random_seed'] = schema.get('random_seed', None) if kwargs.get('random_seed', None) is None else kwargs['random_seed']
        schema['central_agent'] = parse_bool(
            kwargs['central_agent'] if kwargs.get('central_agent') is not None else schema['central_agent'],
            default=False,
            path='central_agent',
        )

        schema['chargers_observations_helper'] = {key: value for key, value in schema["observations"].items() if "electric_vehicle_" in key}
        schema['chargers_actions_helper'] = {key: value for key, value in schema["actions"].items() if "electric_vehicle_" in key}
        schema['chargers_shared_observations_helper'] = {
            key: value
            for key, value in schema["observations"].items()
            if "electric_vehicle_" in key and value.get("shared_in_central_agent", True)
        }

        schema['washing_machine_observations_helper'] = {key: value for key, value in schema["observations"].items() if "washing_machine_" in key}
        schema['washing_machine_actions_helper'] = {key: value for key, value in schema["actions"].items() if "washing_machine" in key}

        schema['observations'] = {
            key: value
            for key, value in schema["observations"].items()
            if key not in set(schema['chargers_observations_helper']) | set(schema['washing_machine_observations_helper'])
        }
        schema['actions'] = {
            key: value
            for key, value in schema['actions'].items()
            if key not in set(schema['chargers_actions_helper']) | set(schema['washing_machine_actions_helper'])
        }

        schema['shared_observations'] = (
            kwargs['shared_observations']
            if kwargs.get('shared_observations') is not None
            else [
                k
                for k, v in schema['observations'].items()
                if not k.startswith("electric_vehicle_")
                and "washing_machine" not in k
                and parse_bool(v.get('shared_in_central_agent', False), default=False, path=f'observations.{k}.shared_in_central_agent')
            ]
        )

        schema['episode_time_steps'] = kwargs['episode_time_steps'] if kwargs.get('episode_time_steps') is not None else schema.get('episode_time_steps', None)
        schema['rolling_episode_split'] = kwargs['rolling_episode_split'] if kwargs.get('rolling_episode_split') is not None else schema.get('rolling_episode_split', None)
        schema['random_episode_split'] = kwargs['random_episode_split'] if kwargs.get('random_episode_split') is not None else schema.get('random_episode_split', None)
        schema['seconds_per_time_step'] = kwargs['seconds_per_time_step'] if kwargs.get('seconds_per_time_step') is not None else schema['seconds_per_time_step']

        schema['simulation_start_time_step'] = kwargs['simulation_start_time_step'] if kwargs.get('simulation_start_time_step') is not None else schema['simulation_start_time_step']
        schema['simulation_end_time_step'] = kwargs['simulation_end_time_step'] if kwargs.get('simulation_end_time_step') is not None else schema['simulation_end_time_step']
        episode_tracker = EpisodeTracker(schema['simulation_start_time_step'], schema['simulation_end_time_step'])

        dataset = DataSet()
        pv_sizing_data = dataset.get_pv_sizing_data()
        battery_sizing_data = dataset.get_battery_sizing_data()

        buildings_to_include = list(schema['buildings'].keys())
        buildings: List[Building] = []

        if kwargs.get('buildings') is not None and len(kwargs['buildings']) > 0:
            if isinstance(kwargs['buildings'][0], Building):
                buildings = kwargs['buildings']

                for building in buildings:
                    building.episode_tracker = episode_tracker

                buildings_to_include = []

            elif isinstance(kwargs['buildings'][0], str):
                buildings_to_include = [b for b in buildings_to_include if b in kwargs['buildings']]

            elif isinstance(kwargs['buildings'][0], int):
                buildings_to_include = [buildings_to_include[i] for i in kwargs['buildings']]

            else:
                raise Exception('Unknown buildings type. Allowed types are citylearn.building.Building, int and str.')

        else:
            buildings_to_include = [
                b for b in buildings_to_include
                if parse_bool(schema['buildings'][b].get('include', True), default=True, path=f'buildings.{b}.include')
            ]

        for i, building_name in enumerate(buildings_to_include):
            buildings.append(self.load_building(i, building_name, schema, episode_tracker, pv_sizing_data, battery_sizing_data, **kwargs))

        electric_vehicles: List[ElectricVehicle] = []
        if kwargs.get('electric_vehicles_def') is not None and len(kwargs['electric_vehicles_def']) > 0:
            electric_vehicle_schemas = kwargs['electric_vehicles_def']
        else:
            electric_vehicle_schemas = schema.get('electric_vehicles_def', {})

        for electric_vehicle_name, electric_vehicle_schema in electric_vehicle_schemas.items():
            if parse_bool(electric_vehicle_schema.get('include', True), default=True, path=f'electric_vehicles_def.{electric_vehicle_name}.include'):
                time_step_ratio = buildings[0].time_step_ratio if len(buildings) > 0 else 1.0
                electric_vehicles.append(
                    self.load_electric_vehicle(electric_vehicle_name, schema, electric_vehicle_schema, episode_tracker, time_step_ratio)
                )

        reward_schema = schema['reward_function']
        reward_type = reward_schema['type']
        reward_attrs = reward_schema.get('attributes', {})
        is_multi = isinstance(reward_type, dict)

        if is_multi:
            default_type = reward_type.get('default')
            if default_type is None and reward_type:
                default_type = next(iter(reward_type.values()))

            default_attrs = reward_attrs.get('default')
            if default_attrs is None and reward_attrs:
                default_attrs = next(iter(reward_attrs.values()))

            reward_functions = {}
            for building in buildings:
                name = building.name
                r_type = reward_type.get(name, default_type)
                r_attr = reward_attrs.get(name, default_attrs) or {}

                if r_type is None:
                    raise ValueError(f"No reward function defined for building '{name}' and no default provided")

                module_name = '.'.join(r_type.split('.')[:-1])
                class_name = r_type.split('.')[-1]
                module = importlib.import_module(module_name)
                constructor = getattr(module, class_name)
                reward_functions[name] = constructor(None, **r_attr)

            reward_function = MultiBuildingRewardFunction(None, reward_functions)

        else:
            if 'reward_function' in kwargs and kwargs['reward_function'] is not None:
                reward_function_type = kwargs['reward_function']
                if not isinstance(reward_function_type, str):
                    reward_function_type = f"{reward_function_type.__module__}.{reward_function_type.__name__}"
            else:
                reward_function_type = reward_type

            reward_function_attributes = kwargs.get('reward_function_kwargs') or reward_attrs or {}

            module_name = '.'.join(reward_function_type.split('.')[:-1])
            class_name = reward_function_type.split('.')[-1]
            module = importlib.import_module(module_name)
            constructor = getattr(module, class_name)
            reward_function = constructor(None, **reward_function_attributes)

        return (
            schema['root_directory'],
            buildings,
            electric_vehicles,
            schema['episode_time_steps'],
            schema['rolling_episode_split'],
            schema['random_episode_split'],
            schema['seconds_per_time_step'],
            reward_function,
            schema['central_agent'],
            schema['shared_observations'],
            episode_tracker,
        )

    def load_building(
        self,
        index: int,
        building_name: str,
        schema: dict,
        episode_tracker: EpisodeTracker,
        pv_sizing_data: pd.DataFrame,
        battery_sizing_data: pd.DataFrame,
        **kwargs,
    ) -> Building:
        """Initialize and return a building model."""

        building_schema = schema['buildings'][building_name]
        building_kwargs = {}
        if building_schema.get('charging_constraints') is not None:
            building_kwargs['charging_constraints'] = building_schema['charging_constraints']
        if building_schema.get('electrical_service') is not None:
            building_kwargs['electrical_service'] = building_schema['electrical_service']
        if building_schema.get('equity_group') is not None:
            building_kwargs['equity_group'] = building_schema.get('equity_group')
        electrical_storage_attributes = (building_schema.get('electrical_storage') or {}).get('attributes', {}) or {}
        if electrical_storage_attributes.get('phase_connection') is not None:
            building_kwargs['electrical_storage_phase_connection'] = electrical_storage_attributes.get('phase_connection')
        seconds_per_time_step = schema['seconds_per_time_step']
        noise_std = building_schema.get('noise_std', 0.0)

        energy_simulation = pd.read_csv(os.path.join(schema['root_directory'], building_schema['energy_simulation']))
        energy_simulation = EnergySimulation(**energy_simulation.to_dict('list'), seconds_per_time_step=seconds_per_time_step, noise_std=noise_std)
        ratios = getattr(energy_simulation, 'time_step_ratios', None) or []
        building_kwargs['time_step_ratio'] = ratios[-1] if len(ratios) > 0 else 1.0
        weather = pd.read_csv(os.path.join(schema['root_directory'], building_schema['weather']))
        weather = Weather(**weather.to_dict('list'), noise_std=noise_std)

        if building_schema.get('carbon_intensity', None) is not None:
            carbon_intensity = pd.read_csv(os.path.join(schema['root_directory'], building_schema['carbon_intensity']))
            carbon_intensity = CarbonIntensity(**carbon_intensity.to_dict('list'), noise_std=noise_std)
        else:
            carbon_intensity = CarbonIntensity(np.zeros(energy_simulation.hour.shape[0], dtype='float32'), noise_std=noise_std)

        if building_schema.get('pricing', None) is not None:
            pricing = pd.read_csv(os.path.join(schema['root_directory'], building_schema['pricing']))
            pricing = Pricing(**pricing.to_dict('list'), noise_std=noise_std)
        else:
            pricing = Pricing(
                np.zeros(energy_simulation.hour.shape[0], dtype='float32'),
                np.zeros(energy_simulation.hour.shape[0], dtype='float32'),
                np.zeros(energy_simulation.hour.shape[0], dtype='float32'),
                np.zeros(energy_simulation.hour.shape[0], dtype='float32'),
                noise_std=noise_std,
            )

        building_type = 'citylearn.citylearn.Building' if building_schema.get('type', None) is None else building_schema['type']
        building_type_module = '.'.join(building_type.split('.')[0:-1])
        building_type_name = building_type.split('.')[-1]
        building_constructor = getattr(importlib.import_module(building_type_module), building_type_name)

        if building_schema.get('dynamics', None) is not None:
            dynamics_type = building_schema['dynamics']['type']
            dynamics_module = '.'.join(dynamics_type.split('.')[0:-1])
            dynamics_name = dynamics_type.split('.')[-1]
            dynamics_constructor = getattr(importlib.import_module(dynamics_module), dynamics_name)
            attributes = building_schema['dynamics'].get('attributes', {})
            attributes['filepath'] = os.path.join(schema['root_directory'], attributes['filename'])
            _ = attributes.pop('filename')
            building_kwargs['dynamics'] = dynamics_constructor(**attributes)
        else:
            building_kwargs['dynamics'] = None

        if building_schema.get('occupant', None) is not None:
            building_occupant = building_schema['occupant']
            occupant_type = building_occupant['type']
            occupant_module = '.'.join(occupant_type.split('.')[0:-1])
            occupant_name = occupant_type.split('.')[-1]
            occupant_constructor = getattr(importlib.import_module(occupant_module), occupant_name)
            attributes: dict = building_occupant.get('attributes', {})
            parameters_filepath = os.path.join(schema['root_directory'], building_occupant['parameters_filename'])
            parameters = pd.read_csv(parameters_filepath)
            attributes['parameters'] = LogisticRegressionOccupantParameters(**parameters.to_dict('list'))
            attributes['episode_tracker'] = episode_tracker
            attributes['random_seed'] = schema['random_seed']

            for key in ['increase', 'decrease']:
                attributes[f'setpoint_{key}_model_filepath'] = os.path.join(schema['root_directory'], attributes[f'setpoint_{key}_model_filename'])
                _ = attributes.pop(f'setpoint_{key}_model_filename')

            building_kwargs['occupant'] = occupant_constructor(**attributes)
        else:
            building_kwargs['occupant'] = None

        building_schema_power_outage = building_schema.get('power_outage', {})
        simulate_power_outage = kwargs.get('simulate_power_outage')
        simulate_power_outage = building_schema_power_outage.get('simulate_power_outage') if simulate_power_outage is None else simulate_power_outage
        simulate_power_outage = simulate_power_outage[index] if isinstance(simulate_power_outage, list) else simulate_power_outage
        stochastic_power_outage = building_schema_power_outage.get('stochastic_power_outage')

        if building_schema_power_outage.get('stochastic_power_outage_model', None) is not None:
            stochastic_power_outage_model_type = building_schema_power_outage['stochastic_power_outage_model']['type']
            stochastic_power_outage_model_module = '.'.join(stochastic_power_outage_model_type.split('.')[0:-1])
            stochastic_power_outage_model_name = stochastic_power_outage_model_type.split('.')[-1]
            stochastic_power_outage_model_constructor = getattr(
                importlib.import_module(stochastic_power_outage_model_module),
                stochastic_power_outage_model_name,
            )
            attributes = building_schema_power_outage.get('stochastic_power_outage_model', {}).get('attributes', {})
            stochastic_power_outage_model = stochastic_power_outage_model_constructor(**attributes)
        else:
            stochastic_power_outage_model = None

        chargers_list = []
        if building_schema.get('chargers', None) is not None:
            for charger_name, charger_config in building_schema['chargers'].items():
                noise_std = charger_config.get('noise_std', 0.0)

                charger_simulation_file = pd.read_csv(
                    os.path.join(schema['root_directory'], charger_config['charger_simulation'])
                ).iloc[schema['simulation_start_time_step']:schema['simulation_end_time_step'] + 1].copy()

                charger_simulation = ChargerSimulation(*charger_simulation_file.values.T, noise_std=noise_std)
                if 'electric_vehicle_current_soc' in charger_simulation_file.columns:
                    current_soc_raw = pd.to_numeric(charger_simulation_file['electric_vehicle_current_soc'], errors='coerce').to_numpy(dtype='float32')
                    current_soc = np.full(current_soc_raw.shape[0], -0.1, dtype='float32')
                    valid = ~np.isnan(current_soc_raw)

                    if np.any(valid):
                        normalized = current_soc_raw[valid]
                        normalized = np.where(normalized > 1.0, normalized / 100.0, normalized)
                        normalized = np.clip(normalized, 0.0, 1.0)
                        current_soc[valid] = normalized.astype('float32')

                    charger_simulation.electric_vehicle_current_soc = current_soc

                charger_type = charger_config['type']
                charger_module = '.'.join(charger_type.split('.')[0:-1])
                charger_class_name = charger_type.split('.')[-1]
                charger_class = getattr(importlib.import_module(charger_module), charger_class_name)
                charger_attributes = dict(charger_config.get('attributes', {}) or {})
                charger_attributes['episode_tracker'] = episode_tracker
                charger_object = charger_class(
                    charger_simulation=charger_simulation,
                    charger_id=charger_name,
                    **charger_attributes,
                    seconds_per_time_step=schema['seconds_per_time_step'],
                    time_step_ratio=building_kwargs['time_step_ratio'],
                )
                chargers_list.append(charger_object)

        washing_machines_list = []
        if kwargs.get('washing_machines') is not None and len(kwargs['washing_machines']) > 0:
            washing_machine_schemas = kwargs['washing_machines']
        else:
            washing_machine_schemas = building_schema.get('washing_machines', {})

        for washing_machine_name, washing_machine_schema in washing_machine_schemas.items():
            washing_machines_list.append(self.load_washing_machine(washing_machine_name, schema, washing_machine_schema, episode_tracker))

        observation_metadata, action_metadata = self.process_metadata(
            schema,
            building_schema,
            chargers_list,
            washing_machines_list,
            index,
            energy_simulation,
            **kwargs,
        )

        building: Building = building_constructor(
            energy_simulation=energy_simulation,
            washing_machines=washing_machines_list,
            electric_vehicle_chargers=chargers_list,
            weather=weather,
            observation_metadata=observation_metadata,
            action_metadata=action_metadata,
            carbon_intensity=carbon_intensity,
            pricing=pricing,
            name=building_name,
            seconds_per_time_step=schema['seconds_per_time_step'],
            random_seed=schema['random_seed'],
            episode_tracker=episode_tracker,
            simulate_power_outage=simulate_power_outage,
            stochastic_power_outage=stochastic_power_outage,
            stochastic_power_outage_model=stochastic_power_outage_model,
            **building_kwargs,
        )

        device_metadata = {
            'cooling_device': {'autosizer': building.autosize_cooling_device},
            'heating_device': {'autosizer': building.autosize_heating_device},
            'dhw_device': {'autosizer': building.autosize_dhw_device},
            'dhw_storage': {'autosizer': building.autosize_dhw_storage},
            'cooling_storage': {'autosizer': building.autosize_cooling_storage},
            'heating_storage': {'autosizer': building.autosize_heating_storage},
            'electrical_storage': {'autosizer': building.autosize_electrical_storage},
            'washing_machine': {'autosizer': building.autosize_electrical_storage},
            'pv': {'autosizer': building.autosize_pv},
        }
        solar_generation = kwargs.get('solar_generation')
        solar_generation = True if solar_generation is None else solar_generation
        solar_generation = solar_generation[index] if isinstance(solar_generation, list) else solar_generation

        for device_name in device_metadata:
            if building_schema.get(device_name, None) is None:
                device = None

            elif device_name == 'pv' and not solar_generation:
                device = None

            else:
                device_type: str = building_schema[device_name]['type']
                device_module = '.'.join(device_type.split('.')[0:-1])
                device_type_name = device_type.split('.')[-1]
                constructor = getattr(importlib.import_module(device_module), device_type_name)
                attributes = dict(building_schema[device_name].get('attributes', {}) or {})
                if device_name == 'electrical_storage':
                    attributes.pop('phase_connection', None)
                attributes['seconds_per_time_step'] = schema['seconds_per_time_step']

                md5 = hashlib.md5()
                device_random_seed = 0

                for string in [building_name, building_type, device_name, device_type]:
                    md5.update(string.encode())
                    hash_to_integer_base = 16
                    device_random_seed += int(md5.hexdigest(), hash_to_integer_base)

                device_random_seed = int(str(device_random_seed * (schema['random_seed'] + 1))[:9])

                attributes = {
                    **attributes,
                    'random_seed': attributes['random_seed'] if attributes.get('random_seed', None) is not None else device_random_seed,
                }
                device = constructor(**attributes)
                autosize = parse_bool(
                    building_schema[device_name].get('autosize', False),
                    default=False,
                    path=f'buildings.{building.name}.{device_name}.autosize',
                )
                building.__setattr__(device_name, device)

                if autosize:
                    autosizer = device_metadata[device_name]['autosizer']
                    autosize_kwargs = {} if building_schema[device_name].get('autosize_attributes', None) is None else building_schema[device_name]['autosize_attributes']

                    if isinstance(device, PV):
                        autosize_kwargs['epw_filepath'] = os.path.join(schema['root_directory'], autosize_kwargs['epw_filepath'])
                        autosize_kwargs['sizing_data'] = pv_sizing_data

                    elif isinstance(device, Battery):
                        autosize_kwargs['sizing_data'] = battery_sizing_data

                    autosizer(**autosize_kwargs)

                device.random_seed = schema['random_seed']

        building.observation_space = building.estimate_observation_space()
        building.action_space = building.estimate_action_space()

        return building

    def process_metadata(
        self,
        schema,
        building_schema,
        chargers_list,
        washing_machines_list,
        index,
        energy_simulation: EnergySimulation,
        **kwargs,
    ):
        """Build observation and action metadata for one building."""

        observation_metadata = {
            k: parse_bool(v.get('active', False), default=False, path=f'observations.{k}.active')
            for k, v in schema['observations'].items()
        }
        if 'minutes' in observation_metadata and energy_simulation.minutes is None:
            observation_metadata.pop('minutes', None)

        chargers_observations_metadata_helper = {
            k: parse_bool(v.get('active', False), default=False, path=f'observations.{k}.active')
            for k, v in schema['chargers_observations_helper'].items()
        }
        washing_machine_observations_metadata_helper = {
            k: parse_bool(v.get('active', False), default=False, path=f'observations.{k}.active')
            for k, v in schema['washing_machine_observations_helper'].items()
        }

        if kwargs.get('active_observations') is not None:
            active_observations = kwargs['active_observations']
            active_observations = active_observations[index] if isinstance(active_observations[0], list) else active_observations
            observation_metadata = {k: True if k in active_observations else False for k in observation_metadata}
            chargers_observations_metadata_helper = {k: True if k in active_observations else False for k in chargers_observations_metadata_helper}
            washing_machine_observations_metadata_helper = {k: True if k in active_observations else False for k in washing_machine_observations_metadata_helper}

        if kwargs.get('inactive_observations') is not None:
            inactive_observations = kwargs['inactive_observations']
            inactive_observations = inactive_observations[index] if isinstance(inactive_observations[0], list) else inactive_observations
        elif building_schema.get('inactive_observations') is not None:
            inactive_observations = building_schema['inactive_observations']
        else:
            inactive_observations = []

        observation_metadata = {
            k: False if k in inactive_observations else observation_metadata[k]
            for k in observation_metadata
        }
        chargers_observations_metadata_helper = {
            k: False if k in inactive_observations else chargers_observations_metadata_helper[k]
            for k in chargers_observations_metadata_helper
        }
        washing_machine_observations_metadata_helper = {
            k: False if k in inactive_observations else washing_machine_observations_metadata_helper[k]
            for k in washing_machine_observations_metadata_helper
        }

        action_metadata = {
            k: parse_bool(v.get('active', False), default=False, path=f'actions.{k}.active')
            for k, v in schema['actions'].items()
        }
        chargers_actions_metadata_helper = {
            k: parse_bool(v.get('active', False), default=False, path=f'actions.{k}.active')
            for k, v in schema['chargers_actions_helper'].items()
        }
        washing_machine_actions_metadata_helper = {
            k: parse_bool(v.get('active', False), default=False, path=f'actions.{k}.active')
            for k, v in schema['washing_machine_actions_helper'].items()
        }

        if kwargs.get('active_actions') is not None:
            active_actions = kwargs['active_actions']
            active_actions = active_actions[index] if isinstance(active_actions[0], list) else active_actions
            action_metadata = {k: True if k in active_actions else False for k in action_metadata}
            chargers_actions_metadata_helper = {k: True if k in active_actions else False for k in chargers_actions_metadata_helper}
            washing_machine_actions_metadata_helper = {k: True if k in active_actions else False for k in washing_machine_actions_metadata_helper}

        if kwargs.get('inactive_actions') is not None:
            inactive_actions = kwargs['inactive_actions']
            inactive_actions = inactive_actions[index] if isinstance(inactive_actions[0], list) else inactive_actions
        elif building_schema.get('inactive_actions') is not None:
            inactive_actions = building_schema['inactive_actions']
        else:
            inactive_actions = []

        action_metadata = {k: False if k in inactive_actions else v for k, v in action_metadata.items()}
        chargers_actions_metadata_helper = {k: False if k in inactive_actions else v for k, v in chargers_actions_metadata_helper.items()}
        washing_machine_actions_metadata_helper = {k: False if k in inactive_actions else v for k, v in washing_machine_actions_metadata_helper.items()}

        if len(chargers_list) > 0:
            for charger in chargers_list:
                charger_id = charger.charger_id

                if chargers_observations_metadata_helper.get('electric_vehicle_charger_connected_state', False):
                    observation_metadata[f'electric_vehicle_charger_{charger_id}_connected_state'] = True

                if chargers_observations_metadata_helper.get('connected_electric_vehicle_at_charger_departure_time', False):
                    observation_metadata[f'connected_electric_vehicle_at_charger_{charger_id}_departure_time'] = True

                if chargers_observations_metadata_helper.get('connected_electric_vehicle_at_charger_required_soc_departure', False):
                    observation_metadata[f'connected_electric_vehicle_at_charger_{charger_id}_required_soc_departure'] = True

                if chargers_observations_metadata_helper.get('connected_electric_vehicle_at_charger_soc', False):
                    observation_metadata[f'connected_electric_vehicle_at_charger_{charger_id}_soc'] = True

                if chargers_observations_metadata_helper.get('connected_electric_vehicle_at_charger_battery_capacity', False):
                    observation_metadata[f'connected_electric_vehicle_at_charger_{charger_id}_battery_capacity'] = True

                if chargers_observations_metadata_helper.get('electric_vehicle_charger_incoming_state', False):
                    observation_metadata[f'electric_vehicle_charger_{charger_id}_incoming_state'] = True

                if chargers_observations_metadata_helper.get('incoming_electric_vehicle_at_charger_estimated_arrival_time', False):
                    observation_metadata[f'incoming_electric_vehicle_at_charger_{charger_id}_estimated_arrival_time'] = True

                if chargers_observations_metadata_helper.get('incoming_electric_vehicle_at_charger_estimated_soc_arrival', False):
                    observation_metadata[f'incoming_electric_vehicle_at_charger_{charger_id}_estimated_soc_arrival'] = True

                if chargers_actions_metadata_helper.get('electric_vehicle_storage', False):
                    action_metadata[f'electric_vehicle_storage_{charger.charger_id}'] = True

        if len(washing_machines_list) > 0:
            for washing_machine in washing_machines_list:
                washing_machine_name = washing_machine.name
                if washing_machine_observations_metadata_helper.get('washing_machine_start_time_step', False):
                    observation_metadata[f'{washing_machine_name}_start_time_step'] = True

                if washing_machine_observations_metadata_helper.get('washing_machine_end_time_step', False):
                    observation_metadata[f'{washing_machine_name}_end_time_step'] = True

                if washing_machine_actions_metadata_helper.get('washing_machine', False):
                    action_metadata[f'{washing_machine_name}'] = True

        return observation_metadata, action_metadata

    def load_electric_vehicle(
        self,
        electric_vehicle_name: str,
        schema: dict,
        electric_vehicle_schema: dict,
        episode_tracker: EpisodeTracker,
        time_step_ratio,
    ) -> ElectricVehicle:
        """Initialize and return an electric vehicle model."""

        capacity = electric_vehicle_schema['battery']['attributes']['capacity']
        nominal_power = electric_vehicle_schema['battery']['attributes']['nominal_power']
        initial_soc = electric_vehicle_schema['battery']['attributes'].get('initial_soc')
        if initial_soc is None:
            seed_source = f"{schema['random_seed']}:{electric_vehicle_name}:initial_soc"
            deterministic_seed = int(hashlib.md5(seed_source.encode('utf-8')).hexdigest()[:8], 16)
            initial_soc = float(np.random.RandomState(deterministic_seed).uniform(0.0, 1.0))
        depth_of_discharge = electric_vehicle_schema['battery']['attributes'].get('depth_of_discharge', 0.10)

        battery = Battery(
            capacity=capacity,
            nominal_power=nominal_power,
            initial_soc=initial_soc,
            seconds_per_time_step=schema['seconds_per_time_step'],
            time_step_ratio=time_step_ratio,
            random_seed=schema['random_seed'],
            episode_tracker=episode_tracker,
            depth_of_discharge=depth_of_discharge,
        )

        electric_vehicle_type = 'citylearn.citylearn.ElectricVehicle' if electric_vehicle_schema.get('type', None) is None else electric_vehicle_schema['type']
        electric_vehicle_type_module = '.'.join(electric_vehicle_type.split('.')[0:-1])
        electric_vehicle_type_name = electric_vehicle_type.split('.')[-1]
        electric_vehicle_constructor = getattr(importlib.import_module(electric_vehicle_type_module), electric_vehicle_type_name)

        electric_vehicle: ElectricVehicle = electric_vehicle_constructor(
            battery=battery,
            name=electric_vehicle_name,
            seconds_per_time_step=schema['seconds_per_time_step'],
            random_seed=schema['random_seed'],
            episode_tracker=episode_tracker,
        )

        return electric_vehicle

    def load_washing_machine(
        self,
        washing_machine_name: str,
        schema: dict,
        washing_machine_schema: dict,
        episode_tracker: EpisodeTracker,
    ) -> WashingMachine:
        """Load simulation data and initialize a `WashingMachine` instance."""

        file_path = os.path.join(schema['root_directory'], washing_machine_schema['washing_machine_energy_simulation'])

        washing_machine_simulation = pd.read_csv(file_path).iloc[
            schema['simulation_start_time_step']:schema['simulation_end_time_step'] + 1
        ].copy()

        washing_machine_simulation = WashingMachineSimulation(*washing_machine_simulation.values.T)

        washing_machine = WashingMachine(
            washing_machine_simulation=washing_machine_simulation,
            episode_tracker=episode_tracker,
            name=washing_machine_name,
            seconds_per_time_step=schema['seconds_per_time_step'],
            random_seed=schema['random_seed'],
        )

        return washing_machine



import math
from typing import Any, List, Tuple
import numpy as np
from citylearn.agents.base import Agent
from citylearn.citylearn import CityLearnEnv

class TabularQLearning(Agent):
    """Implementation of Tabular Q-Learning algorithm for discrete observation and action space and epsilon-greedy action selection.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    epsilon: float, default: 1.0
        Exploration rate.
    minimum_epsilon: float, default: 0.01
        Minimum value exporation rate can decay to.
    epsilon_decay: float, default: 0.0001
        :code:`epsilon` exponential decay rate.
    learning_rate: float, default: 0.05
        Defines to what degree new knowledge overrides old knowledge: for :code:`learning_rate` = 0, no
        learning happens, while for :code:`learning_rate` = 1, all prior knowledge is lost.
    discount_factor: float, default: 0.90
        Balance between an agent that considers only immediate rewards (:code:`discount_factor` = 0) and
        one that strives towards long term rewards (:code:`discount_factor` = 1)
    q_init_value: float, default: np.nan
        Q-Table initialization value.

    Other Parameters
    ----------------
    **kwargs: Any
        Other keyword arguments used to initialize :py:class:`citylearn.agents.base.Agent` super class.
    """
    
    def __init__(
        self, env: CityLearnEnv, epsilon: float = None, minimum_epsilon: float = None, epsilon_decay: float = None, 
        learning_rate: float = None, discount_factor: float = None, q_init_value: float = None, **kwargs: Any,
    ):
        super().__init__(env, **kwargs)
        self.epsilon = 1.0 if epsilon is None else epsilon
        self.epsilon_init = self.epsilon
        self.minimum_epsilon = 0.01 if minimum_epsilon is None else minimum_epsilon
        self.epsilon_decay = 0.0001 if epsilon_decay is None else epsilon_decay
        self.learning_rate = 0.05 if learning_rate is None else learning_rate
        self.discount_factor = 0.90 if discount_factor is None else discount_factor
        self.q_init_value = np.nan if q_init_value is None else q_init_value
        self.q, self.q_exploration, self.q_exploitation = self.__initialize_q()
        self.__explored = False

    def predict(self, observations: List[List[float]], deterministic: bool = None) -> List[List[float]]:
        """Provide actions for current time step.

        If `deterministic` = True or, randomly generated number is greater than `epsilon`, return
        deterministic action from Q-Table i.e. action with max Q-value for given observations 
        otherwise, return randomly sampled action.
        
        Parameters
        ----------
        observations: List[List[float]]
            Environment observations
        deterministic: bool, default: False
            Wether to return purely exploitatative deterministic actions.

        Returns
        -------
        actions: List[List[float]]
            Action values
        """
        
        deterministic = False if deterministic is None else deterministic
        actions = None
        seed = self.random_seed if self.random_seed is None else self.random_seed + self.time_step
        nprs = np.random.RandomState(seed)
        

        if deterministic or nprs.random() > self.epsilon:
            # Use q-function to decide action
            actions = self.__exploit(observations)
            self.__explored = False
        
        else:
            # Explore random action
            actions = [[s.sample()] for s in self.action_space]
            self.__explored = True

        # exponential decay
        episode = int(self.time_step/self.episode_time_steps)
        self.epsilon = max(self.minimum_epsilon, self.epsilon_init*np.exp(-self.epsilon_decay*episode))

        self.actions = actions
        self.next_time_step()
        return actions

    def __exploit(self, observations: List[List[float]]) -> List[List[float]]:
        """Select deterministic actions from Q-Table i.e. action with max Q-value for given observations."""
        
        actions = []

        for i, o in enumerate(observations):
            o = o[0]

            try:
                a = np.nanargmax(self.q[i][o])
            
            except ValueError:
                # when all values for observation are still NaN
                a = self.action_space[i].sample()

            actions.append([a])
        
        return actions

    def update(self, observations: List[List[float]], actions: List[List[float]], reward: List[float], next_observations: List[List[float]], terminated: bool, truncated: bool):
        r"""Update Q-Table using Bellman equation.

        Parameters
        ----------
        observations : List[List[float]]
            Previous time step observations.
        actions : List[List[float]]
            Previous time step actions.
        reward : List[float]
            Current time step reward.
        next_observations : List[List[float]]
            Current time step observations.
        terminated : bool
            Indication that episode has ended.
        truncated : bool
            If episode truncates due to a time limit or a reason that is not defined as part of the task MDP.
        """

        # Compute temporal difference target and error to udpate q-function
        
        for i, (o, a, r, n) in enumerate(zip(observations, actions, reward, next_observations)):
            o, n, a = o[0], n[0], a[0]
            current_q = self.q[i][o, a]
            current_q = 0.0 if math.isnan(current_q) else current_q

            try:
                next_max_q = np.nanargmax(self.q[i][n])
            
            except ValueError:
                # when all values for observation are still NaN
                next_max_q = 0.0
            
            # update q
            new_q = current_q + self.learning_rate*(r + self.discount_factor*next_max_q - current_q)
            self.q[i][o, a] = new_q

            # update exploration-exploitation count
            if self.__explored:
                self.q_exploration[i][o, a] += 1
            else:
                self.q_exploitation[i][o, a] += 1

    def __initialize_q(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Initialize all values in Q-Table with `q_init_value`."""
        
        q = [None for _ in self.observation_space]
        q_exploration = [None for _ in self.observation_space]
        q_exploitation = [None for _ in self.observation_space]

        for i, (od, ad) in enumerate(zip(self.observation_space, self.action_space)):
            shape = (od.n, ad.n)
            q[i] = np.ones(shape=shape)*self.q_init_value
            q_exploration[i] = np.zeros(shape=shape)
            q_exploitation[i] = np.zeros(shape=shape)
        
        return q, q_exploration, q_exploitation


import json
from pathlib import Path

import numpy as np
import pytest

pytest.importorskip("gymnasium")

from citylearn.agents.rbc import BasicElectricVehicleRBC_ReferenceController as Agent
from citylearn.citylearn import CityLearnEnv
from citylearn.internal.runtime import CityLearnRuntimeService


SCHEMA = Path(__file__).resolve().parents[1] / "data/datasets/citylearn_challenge_2022_phase_all_plus_evs/schema.json"


def test_agent_learn_handles_single_timestep_episode_without_calling_step(monkeypatch):
    env = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=1, render_mode="none", random_seed=7)
    agent = Agent(env)

    step_called = False

    def _unexpected_step(_actions):
        nonlocal step_called
        step_called = True
        raise AssertionError("env.step() should not be called when episode is terminal after reset.")

    monkeypatch.setattr(env, "step", _unexpected_step)

    try:
        agent.learn(episodes=1, deterministic=True, logging_level=40)
        assert not step_called
    finally:
        env.close()


def _run_rbc_episode(render_mode: str, output_root: Path, seed: int = 7):
    env = CityLearnEnv(
        str(SCHEMA),
        central_agent=True,
        episode_time_steps=128,
        render_mode=render_mode,
        render_directory=output_root,
        random_seed=seed,
    )
    agent = Agent(env)

    try:
        observations, _ = env.reset()
        rewards = []

        while not env.terminated:
            actions = agent.predict(observations, deterministic=True)
            observations, reward, terminated, truncated, _ = env.step(actions)
            rewards.append(np.array(reward, dtype="float64").reshape(-1))

            if terminated or truncated:
                break

        reward_trace = np.vstack(rewards) if rewards else np.zeros((0, len(env.action_space)), dtype="float64")
        ev_soc_trace = {
            ev.name: np.array(ev.battery.soc[: env.time_step + 1], dtype="float64")
            for ev in env.electric_vehicles
        }
        return reward_trace, ev_soc_trace
    finally:
        env.close()


def test_same_seed_is_reproducible_across_render_modes_without_global_numpy_seed(tmp_path):
    during_rewards, during_ev_soc = _run_rbc_episode("during", tmp_path / "during")
    end_rewards, end_ev_soc = _run_rbc_episode("end", tmp_path / "end")

    assert during_rewards.shape == end_rewards.shape
    assert np.allclose(during_rewards, end_rewards)
    assert set(during_ev_soc.keys()) == set(end_ev_soc.keys())

    for ev_name in during_ev_soc:
        assert during_ev_soc[ev_name].shape == end_ev_soc[ev_name].shape
        assert np.allclose(during_ev_soc[ev_name], end_ev_soc[ev_name])


def test_runtime_random_seed_overrides_schema_seed_for_loading_defaults():
    schema = json.loads(SCHEMA.read_text(encoding="utf-8"))
    missing_initial_soc = [
        name
        for name, ev in (schema.get("electric_vehicles_def", {}) or {}).items()
        if "initial_soc" not in ((ev.get("battery", {}) or {}).get("attributes", {}) or {})
    ]
    assert missing_initial_soc, "Expected at least one EV without explicit initial_soc in schema."

    env_a = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=8, render_mode="none", random_seed=1)
    env_b = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=8, render_mode="none", random_seed=2)
    env_c = CityLearnEnv(str(SCHEMA), central_agent=True, episode_time_steps=8, render_mode="none", random_seed=1)

    try:
        soc_a = {ev.name: float(ev.battery.initial_soc) for ev in env_a.electric_vehicles}
        soc_b = {ev.name: float(ev.battery.initial_soc) for ev in env_b.electric_vehicles}
        soc_c = {ev.name: float(ev.battery.initial_soc) for ev in env_c.electric_vehicles}

        # different runtime seeds should alter default-initialized EV SOCs
        assert any(not np.isclose(soc_a[name], soc_b[name]) for name in missing_initial_soc)

        # same runtime seed should reproduce the same default SOCs
        assert all(np.isclose(soc_a[name], soc_c[name]) for name in missing_initial_soc)
    finally:
        env_a.close()
        env_b.close()
        env_c.close()


def test_ev_unconnected_drift_std_scales_with_physical_timestep():
    hourly = CityLearnRuntimeService._ev_unconnected_drift_std(3600)
    minute = CityLearnRuntimeService._ev_unconnected_drift_std(60)
    quarter_hour = CityLearnRuntimeService._ev_unconnected_drift_std(900)

    assert np.isclose(hourly, 0.2)
    assert np.isclose(quarter_hour, 0.1)
    assert minute < quarter_hour < hourly


def test_unconnected_ev_soc_drift_uses_time_aware_variance():
    class _RandomState:
        def __init__(self):
            self.calls = []

        def normal(self, loc, scale):
            self.calls.append((float(loc), float(scale)))
            return float(loc + scale)

    class _Battery:
        def __init__(self, soc, target_index):
            self.soc = list(soc)
            self._target_index = target_index

        def force_set_soc(self, value):
            self.soc[self._target_index] = float(value)

    class _EV:
        def __init__(self, name, battery):
            self.name = name
            self.battery = battery

    class _EpisodeTracker:
        def __init__(self, episode_time_steps):
            self.episode_time_steps = episode_time_steps
            self.episode = 0

    class _Env:
        def __init__(self, seconds_per_time_step):
            self.seconds_per_time_step = seconds_per_time_step
            self.time_step = 1
            self.episode_tracker = _EpisodeTracker(episode_time_steps=8)
            self.electric_vehicles = [_EV("EV1", _Battery([0.5, 0.0], target_index=1))]
            self.buildings = []
            self.random_seed = 7
            self._ev_drift_random_state = _RandomState()

    env_hourly = _Env(3600)
    env_minute = _Env(60)
    CityLearnRuntimeService(env_hourly).simulate_unconnected_ev_soc()
    CityLearnRuntimeService(env_minute).simulate_unconnected_ev_soc()

    hourly_scale = env_hourly._ev_drift_random_state.calls[0][1]
    minute_scale = env_minute._ev_drift_random_state.calls[0][1]

    assert hourly_scale > minute_scale
    assert env_hourly.electric_vehicles[0].battery.soc[1] > env_minute.electric_vehicles[0].battery.soc[1]
